import random
from typing import Any, Optional, Union

import abjad

from .. import get, mutate, select


class Shuffler:
    r"""Takes an |abjad.Container| (or child class) and shuffles or rotates its
    logical ties or pitches. When shuffling or rotating pitches only, tuplets
    are supported, otherwise tuplets are not supported.

    Basic usage:
        Calling the object will output a shuffled selection of the input
        container.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> abjad.show(container)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Shuffler-1oLQZeQNPo.png

        >>> shuffler = auxjad.Shuffler(container)
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                d'4
                c'4
                f'4
                e'4
            }

        ..  figure:: ../_images/Shuffler-z2om98675v.png

        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                e'4
                d'4
                f'4
            }

        ..  figure:: ../_images/Shuffler-xu7sln4vt7n.png

        To get the result of the last operation, use the property
        :attr:`current_window`.

        >>> notes = shuffler.current_window
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                e'4
                d'4
                f'4
            }

        ..  figure:: ../_images/Shuffler-gphtpqn9jb.png

        Calling the object outputs the same result as using the method
        :meth:`shuffle`.

        >>> notes = shuffler.shuffle()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                e'4
                f'4
                c'4
                d'4
            }

        ..  figure:: ../_images/Shuffler-g965k0d03if.png

    ..  warning::

        Unlike the other classes in Auxjad, the very first call of an instance
        of this class will already process the initial container. To disable
        this behaviour and output the initial container once before shuffling
        or rotating it, initialise the class with the keyword argument
        :attr:`process_on_first_call` set to ``False``.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> shuffler = auxjad.Shuffler(container,
        ...                            process_on_first_call=False,
        ...                            )
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Shuffler-76039tn5b9k.png

    :func:`len()`:
        Applying the :func:`len()` function to the shuffler will return the
        number of logical ties of :attr:`contents`.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4 ~ | f'2 g'2")
        >>> shuffler = auxjad.Shuffler(container)
        >>> len(shuffler)
        5

        Do note that consecutive rests are considered as a single logical tie,
        so in the example below the :func:`len()` function returns ``5`` and
        not ``6``. When shuffling or rotating logical ties, consecutive rests
        are also shuffled and rotated together.

        >>> container = abjad.Container(r"c'8. d'4 r8 r8. e'16 f'8.")
        >>> shuffler = auxjad.Shuffler(container)
        >>> len(shuffler)
        5

    Arguments and properties:
        This class has many keyword arguments, all of which can be altered
        after instantiation using properties with the same names as shown
        below. Setting :attr:`pitch_only` to ``True`` will enable pitch mode;
        by default, this class shuffles and rotates logical ties, but in pitch
        mode only pitches are shuffled or rotated. By setting
        :attr:`preserve_rest_position` to ``True`` the shuffle and rotation
        operations will not change the position or duration of rests.
        :attr:`disable_rewrite_meter` disables the
        |abjad.Meter.rewrite_meter()| mutation which is applied to the
        container after every call, and :attr:`omit_time_signatures` will
        remove all time signatures from the output (both are ``False`` by
        default). The properties :attr:`boundary_depth`,
        :attr:`maximum_dot_count`, and :attr:`rewrite_tuplets` are passed as
        arguments to |abjad.Meter.rewrite_meter()|, see its documentation
        for more information. By default, calling the object will first return
        the original container and subsequent calls will process it; set
        :attr:`process_on_first_call` to ``True`` and the shuffling process
        will be applied on the very first call.

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'4 d'4 e'4 \time 2/4 f'4 g'4"
        ... )
        >>> shuffler = auxjad.Shuffler(container,
        ...                            pitch_only=False,
        ...                            preserve_rest_position=True,
        ...                            disable_rewrite_meter=False,
        ...                            omit_time_signatures=True,
        ...                            boundary_depth=0,
        ...                            maximum_dot_count=1,
        ...                            rewrite_tuplets=False,
        ...                            process_on_first_call=True,
        ...                            swap_limit=3,
        ...                            )
        >>> shuffler.pitch_only
        False
        >>> shuffler.preserve_rest_position
        True
        >>> shuffler.disable_rewrite_meter
        False
        >>> shuffler.omit_time_signatures
        True
        >>> shuffler.boundary_depth
        0
        >>> shuffler.maximum_dot_count
        1
        >>> shuffler.rewrite_tuplets
        False
        >>> shuffler.process_on_first_call
        True
        >>> shuffler.swap_limit
        3

        Use the properties below to change these values after initialisation.

        >>> shuffler.pitch_only = True
        >>> shuffler.preserve_rest_position = False
        >>> shuffler.disable_rewrite_meter = True
        >>> shuffler.omit_time_signatures = False
        >>> shuffler.boundary_depth = 1
        >>> shuffler.maximum_dot_count = 2
        >>> shuffler.rewrite_tuplets = True
        >>> shuffler.process_on_first_call = False
        >>> shuffler.swap_limit = None
        >>> shuffler.pitch_only
        True
        >>> shuffler.preserve_rest_position
        True
        >>> shuffler.disable_rewrite_meter
        True
        >>> shuffler.omit_time_signatures
        False
        >>> shuffler.boundary_depth
        1
        >>> shuffler.maximum_dot_count
        2
        >>> shuffler.rewrite_tuplets
        True
        >>> shuffler.process_on_first_call
        False
        >>> shuffler.swap_limit
        None

    :attr:`pitch_only`:
        By default, the shuffling operation will shuffle logical ties:

        >>> container = abjad.Container(r"c'8. d'4 r8 r8. e'16 f'8.")
        >>> shuffler = auxjad.Shuffler(container)
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                r4
                r16
                e'16
                f'8
                ~
                f'16
                d'8.
                ~
                d'16
                c'8.
            }

        ..  figure:: ../_images/Shuffler-5j79m0wuxu.png

        Setting :attr:`pitch_only` to ``True`` enables pitch mode, so only
        pitches are shuffled (and not durations). Note how in the example below
        the duration of each leaf is the same as the input container.

        >>> container = abjad.Container(r"c'8. d'4 r8 r8. e'16 f'8.")
        >>> shuffler = auxjad.Shuffler(container, pitch_only=True)
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                f'8.
                r4
                d'8
                ~
                d'8.
                ~
                c'16
                e'8.
            }

        ..  figure:: ../_images/Shuffler-f9jbzqkrkdf.png

    ..  note::

        Altering the value of :attr:`pitch_only`: will replace the original
        :attr:`contents`: with the contents of :attr:`current_window`. Note how
        in the example below, the shuffled leaves in measure 3 comes from the
        previous measure and not from the initial :attr:`contents`:.

        >>> container = abjad.Container(r"c'4.. d'16 e'4. f'8")
        >>> shuffler = auxjad.Shuffler(container, pitch_only=True)
        >>> notes = shuffler.shuffle_n(2)
        >>> staff = abjad.Staff(notes)
        >>> shuffler.pitch_only = False
        >>> notes = shuffler.shuffle_n(2)
        >>> staff.append(notes)
        >>> auxjad.mutate.remove_repeated_time_signatures(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                d'4..
                c'16
                f'4.
                e'8
                d'4..
                f'16
                c'4.
                e'8
                f'16
                d'4..
                e'8
                c'4.
                c'4.
                d'8
                ~
                d'4
                ~
                d'16
                e'8
                f'16
            }

        ..  figure:: ../_images/Shuffler-tyq8y6q8zr9.png

    :attr:`swap_limit`:
        The attribute :attr:`swap_limit` can be used to set the number of times
        that pairs of elements are swapped on a single invocation of the
        shuffling process. Set :attr:`swap_limit` to ``None`` to not limit the
        shuffling process.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> shuffler = auxjad.Shuffler(container,
        ...                            swap_limit=1,
        ...                            )
        >>> notes = shuffler.shuffle_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8
                d'8
                e'8
                g'8
                f'8
                a'8
                b'8
                c''8
                c'8
                c''8
                e'8
                g'8
                f'8
                a'8
                b'8
                d'8
                c'8
                c''8
                e'8
                g'8
                b'8
                a'8
                f'8
                d'8
            }

        ..  figure:: ../_images/Shuffler-Nbo5S6wcfQ.png

    :meth:`rotate`:
        Besides shuffling, logical ties and pitches can also be rotated using
        the :meth:`rotate` method. Similarly to shuffling, it can be applied to
        logical ties or pitches only depending on the property
        :attr:`pitch_only`.

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'16 d'8. ~ d'4 e'4 r4 f'4 ~ f'8.. g'32"
        ... )
        >>> shuffler = auxjad.Shuffler(container)
        >>> notes = shuffler.rotate()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                d'4..
                e'16
                ~
                e'8.
                r16
                r8.
                f'16
                ~
                f'4
                ~
                f'8
                ~
                f'32
                g'32
                c'16
            }

        ..  figure:: ../_images/Shuffler-7vamgsxlr6.png

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'16 d'8. ~ d'4 e'4 r4 f'4 ~ f'8.. g'32"
        ... )
        >>> shuffler = auxjad.Shuffler(container, pitch_only=True)
        >>> notes = shuffler.rotate()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                d'16
                e'8.
                ~
                e'4
                r4
                f'4
                g'4
                ~
                g'8..
                c'32
            }

        ..  figure:: ../_images/Shuffler-89cx79bjji8.png

        This method can also take the optional parameters ``n_rotations`` and
        ``anticlockwise``. The first is an :obj:`int` setting the number of
        rotations applied to the material, and the second is a :obj:`bool`
        setting the direction of the rotation (default ``False``).

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'16 d'8. ~ d'4 e'4 r4 f'4 ~ f'8.. g'32"
        ... )
        >>> shuffler = auxjad.Shuffler(container, pitch_only=True)
        >>> notes = shuffler.rotate(n_rotations=2, anticlockwise=True)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                f'16
                g'8.
                ~
                g'4
                c'4
                d'4
                e'4
                ~
                e'8..
                r32
            }

        ..  figure:: ../_images/Shuffler-g6v6wjm12ub.png

    :attr:`preserve_rest_position`:
        If :attr:`preserve_rest_position` is set to ``True``, the positions of
        all rests will remain the same after either shuffling and rotation. In
        pitch mode (when :attr:`pitch_only` is set to ``True``), this means
        that only the pitched notes will be shuffled or rotated, while the
        rests remain in the exact same place.

        >>> container = abjad.Container(r"c'8. d'4 r8 r8. e'16 f'8.")
        >>> shuffler = auxjad.Shuffler(container,
        ...                            pitch_only=True,
        ...                            preserve_rest_position=True,
        ...                            )
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                d'8.
                f'4
                r8
                r8.
                c'16
                e'8.
            }

        ..  figure:: ../_images/Shuffler-pmou83f7rlj.png

        In logical ties mode, the rests will remain at the same index and will
        have the same total duration as before, but their position in the
        measure might vary since the duration of the pitched logical ties
        preceding it might change.

        >>> container = abjad.Container(r"c'8. d'4 r8 r8. e'16 f'8.")
        >>> shuffler = auxjad.Shuffler(container, preserve_rest_position=True)
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                d'4
                e'16
                r8.
                r8
                f'8
                ~
                f'16
                c'8.
            }

        ..  figure:: ../_images/Shuffler-7hbp2kdpqof.png

    :attr:`disable_rewrite_meter`:
        If :attr:`disable_rewrite_meter` is set to ``True``, then the automatic
        behaviour of rewriting the leaves according to the meter is disabled.

        >>> container = abjad.Container(r"c'4 d'8 e'8 f'2")
        >>> shuffler = auxjad.Shuffler(container,
        ...                            disable_rewrite_meter=True,
        ...                            )
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                e'8
                f'2
                c'4
                d'8
            }

        ..  figure:: ../_images/Shuffler-tb78izpzvjp.png

    :meth:`shuffle_n` and :meth:`rotate_n`:
        To output several shuffled containers at once, use the methods
        :meth:`shuffle_n` and :meth:`rotate_n`, inputting the desired number of
        iterations. :meth:`rotate_n` can also take the optional arguments
        ``n_rotations`` and ``anticlockwise``, similarly to :meth:`rotate`.

        >>> container = abjad.Container(r"c'4 d'8 e'4. f'8. g'16")
        >>> shuffler = auxjad.Shuffler(container)
        >>> notes = shuffler.shuffle_n(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                d'8
                g'16
                c'16
                ~
                c'8.
                f'16
                ~
                f'8
                e'4.
                g'16
                f'8.
                e'4
                ~
                e'8
                c'4
                d'8
            }

        ..  figure:: ../_images/Shuffler-vtia65lbk5.png

        >>> container = abjad.Container(r"c'4 d'8 e'4. f'8. g'16")
        >>> shuffler = auxjad.Shuffler(container)
        >>> notes = shuffler.rotate_n(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                d'8
                e'4.
                f'8.
                g'16
                c'4
                e'4.
                f'8
                ~
                f'16
                g'16
                c'4
                d'8
            }

        ..  figure:: ../_images/Shuffler-3dqhy8eoiez.png

    :attr:`omit_time_signatures`:
        To disable time signatures altogether, initialise this class with the
        keyword argument :attr:`omit_time_signatures` set to ``True`` (default
        is ``False``), or change the :attr:`omit_time_signatures` property
        after initialisation.

        >>> container = abjad.Container(r"\time 3/4 c'16 d'4.. e'4 | r4 f'2")
        >>> shuffler = auxjad.Shuffler(container,
        ...                            omit_time_signatures=True,
        ...                            )
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                d'4..
                e'16
                ~
                e'8.
                f'16
                ~
                f'4..
                r16
                r8.
                c'16
            }

        ..  figure:: ../_images/Shuffler-1v3lwhj430b.png

    ..  tip::

        All methods that return an |abjad.Selection| will add an initial time
        signature to it. The :meth:`shuffle_n` and :meth:`rotate_n` methods
        automatically remove repeated time signatures. When joining selections
        output by multiple method calls, use
        |auxjad.mutate.remove_repeated_time_signatures()| on the whole
        container after fusing the selections to remove any unecessary time
        signature changes.

    Time signature changes:
        This class handles time signature changes too:

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'8. d'4 r8 r8. \time 2/4 e'16 f'4.."
        ... )
        >>> shuffler = auxjad.Shuffler(container)
        >>> notes = shuffler.shuffle_n(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                e'16
                d'8.
                ~
                d'16
                f'4..
                \time 2/4
                c'8.
                r16
                r4
                \time 3/4
                c'8.
                f'16
                ~
                f'4.
                r8
                \time 2/4
                r8.
                d'16
                ~
                d'8.
                e'16
            }

        ..  figure:: ../_images/Shuffler-yx11u6o14v.png

    Tuplet support:
        Tuplets are supported when :attr:`pitch_only` is ``True`` (pitch-only
        mode).

        >>> container = abjad.Container(
        ...     r"\time 5/4 r4 \times 2/3 {c'4 d'2} e'4. f'8"
        ... )
        >>> shuffler = auxjad.Shuffler(container, pitch_only=True)
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 5/4
                d'4
                \times 2/3
                {
                    f'4
                    c'2
                }
                e'4.
                r8
            }

        ..  figure:: ../_images/Shuffler-mjxubkel8y.png

    ..  error::

        Tuplets are not supported when :attr:`pitch_only` is ``False`` (logical
        tie mode). Using a container with tuplets and :attr:`pitch_only` set to
        ``True`` will raise a :exc:`TypeError` exception.

        >>> container = abjad.Container(
        ...     r"\time 5/4 r4 \times 2/3 {c'4 d'2} e'4. f'8"
        ... )
        >>> shuffler = auxjad.Shuffler(container)
        >>> notes = shuffler()
        TypeError: 'contents' contain one ore more tuplets, which are not
        currently supported by the shuffle method

    Indicators:
        This class can also handle dynamics and articulations.

        >>> container = abjad.Container(
        ...     r"<c' e' g'>4--\p d'8-. e'8-. f'4-^\f r4"
        ... )
        >>> shuffler = auxjad.Shuffler(container)
        >>> notes = shuffler.shuffle_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                e'8
                \p
                - \staccato
                d'8
                - \staccato
                f'4
                \f
                - \marcato
                <c' e' g'>4
                \p
                - \tenuto
                r4
                r4
                d'8
                - \staccato
                f'8
                \f
                - \marcato
                ~
                f'8
                <c' e' g'>4
                \p
                - \tenuto
                e'8
                - \staccato
                f'4
                \f
                - \marcato
                e'8
                \p
                - \staccato
                <c' e' g'>8
                - \tenuto
                ~
                <c' e' g'>8
                d'8
                - \staccato
                r4
            }

        ..  figure:: ../_images/Shuffler-2ibui58pj8w.png

    ..  tip::

        The functions |auxjad.mutate.remove_repeated_dynamics()| and
        |auxjad.mutate.reposition_clefs()| can be used to clean the output
        and remove repeated dynamics and unnecessary clef changes.

    ..  warning::

        Do note that some elements that span multiple notes (such as ottava
        indicators, manual beams, etc.) can become problematic when notes
        containing them are split into two. As a rule of thumb, it is always
        better to attach those to the music after the fading process has ended.
        In the case of shuffling logical ties, slurs and hairpins can also
        become a problem, since their start and end position can shift around.
        Dynamics are shuffled together with their leaves, so the initial leaf
        may lack a dynamic marking.

    :attr:`contents`:
        Use the property :attr:`contents` to get the input container upon which
        the shuffler operates. Notice that :attr:`contents` remains invariant
        after any shuffling or rotation operations (use :attr:`current_window`
        for the transformed selection of music). :attr:`contents` can be used
        to change the |abjad.Container| to be shuffled.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> shuffler = auxjad.Shuffler(container)
        >>> abjad.show(shuffler.contents)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Shuffler-qsee7chymo.png

        >>> shuffler()
        >>> abjad.show(shuffler.contents)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Shuffler-ii3fxe001ki.png

        >>> shuffler.contents = abjad.Container(r"cs2 ds2")
        >>> abjad.show(shuffler.contents)

        ..  docs::

            {
                cs2
                ds2
            }

        ..  figure:: ../_images/Shuffler-p2vd4mfvucp.png

    Tweaking |abjad.Meter.rewrite_meter()|:
        This function uses the default logical tie splitting algorithm from
        |abjad.Meter.rewrite_meter()|.

        >>> container = abjad.Container(r"c'4. d'8 e'2")
        >>> shuffler = auxjad.Shuffler(container)
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                e'2
                c'4.
                d'8
            }

        ..  figure:: ../_images/Shuffler-t4lsqxg18ab.png

        Set :attr:`boundary_depth` to a different number to change its
        behaviour.

        >>> shuffler = auxjad.Shuffler(container,
        ...                            boundary_depth=1,
        ...                            )
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                e'2
                c'4
                ~
                c'8
                d'8
            }

        ..  figure:: ../_images/Shuffler-7na5znnhhwe.png

        Other arguments available for tweaking the output of
        |abjad.Meter.rewrite_meter()| are :attr:`maximum_dot_count` and
        :attr:`rewrite_tuplets`, which work exactly as the identically named
        arguments of |abjad.Meter.rewrite_meter()|.

        This class also accepts the arguments ``fuse_across_groups_of_beats``,
        ``fuse_quadruple_meter``, ``fuse_triple_meter``, and
        ``extract_trivial_tuplets``, which are passed on to
        |auxjad.mutate.prettify_rewrite_meter()| (the latter can be disabled
        by setting ``prettify_rewrite_meter`` to ``False``). See the
        documentation of this function for more details on these arguments.

    :attr:`disable_rewrite_meter`:
        By default, this class uses the |abjad.Meter.rewrite_meter()|
        mutation.

        >>> container = abjad.Container(r"c'4 d'8 e'8 f'2")
        >>> shuffler = auxjad.Shuffler(container)
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                e'8
                f'8
                ~
                f'4.
                c'8
                ~
                c'8
                d'8
            }

        ..  figure:: ../_images/Shuffler-7cfnxx7shci.png

        Set :attr:`disable_rewrite_meter` to ``True`` in order to disable this
        behaviour.

        >>> container = abjad.Container(r"c'4 d'8. e'16 f'2")
        >>> abjad.show(container)

        ..  docs::

            {
                e'16
                f'8.
                ~
                f'4
                ~
                f'16
                c'8.
                ~
                c'16
                d'8.
            }

        ..  figure:: ../_images/Shuffler-6gm4ev48j9k.png

        >>> shuffler = auxjad.Shuffler(container,
        ...                            disable_rewrite_meter=True,
        ...                            )
        >>> notes = shuffler()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                e'16
                f'2
                c'4
                d'8.
            }

        ..  figure:: ../_images/Shuffler-xlr4x3bhj6n.png

    Using as iterator:
        The instances of this class can also be used as an iterator, which can
        then be used in a for loop. Note that unlike the methods
        :meth:`shuffle_n` and :meth:`rotate_n`, time signatures are added to
        each window returned by the shuffler. Use the function
        |auxjad.mutate.remove_repeated_time_signatures()| to clean the output
        when using this class in this way. It is also important to note that a
        ``break`` statement is needed when using this class as an iterator. The
        reason is that shuffling is a process that can happen indefinitely
        (unlike some of the other classes in this library).

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> shuffler = auxjad.Shuffler(container)
        >>> staff = abjad.Staff()
        >>> for window in shuffler:
        ...     staff.append(window)
        ...     if abjad.get.duration(staff) == abjad.Duration((9, 4)):
        ...         break
        >>> auxjad.mutate.remove_repeated_time_signatures(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                e'4
                c'4
                d'4
                d'4
                c'4
                e'4
                c'4
                e'4
                d'4
            }

        ..  figure:: ../_images/Shuffler-3gyz7atvemx.png
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_contents",
        "_pitch_only",
        "_preserve_rest_position",
        "_disable_rewrite_meter",
        "_omit_time_signatures",
        "_current_window",
        "_logical_selections",
        "_logical_selections_indeces",
        "_pitches",
        "_time_signatures",
        "_is_first_window",
        "_boundary_depth",
        "_maximum_dot_count",
        "_rewrite_tuplets",
        "_process_on_first_call",
        "_prettify_rewrite_meter",
        "_extract_trivial_tuplets",
        "_fuse_across_groups_of_beats",
        "_fuse_quadruple_meter",
        "_fuse_triple_meter",
        "_swap_limit",
    )

    ### INITIALISER ###

    def __init__(
        self,
        contents: abjad.Container,
        *,
        pitch_only: bool = False,
        preserve_rest_position: bool = False,
        disable_rewrite_meter: bool = False,
        omit_time_signatures: bool = False,
        boundary_depth: Optional[int] = None,
        maximum_dot_count: Optional[int] = None,
        rewrite_tuplets: bool = True,
        process_on_first_call: bool = True,
        prettify_rewrite_meter: bool = True,
        extract_trivial_tuplets: bool = True,
        fuse_across_groups_of_beats: bool = True,
        fuse_quadruple_meter: bool = True,
        fuse_triple_meter: bool = True,
        swap_limit: Optional[int] = None,
    ) -> None:
        self.contents = contents
        self.pitch_only = pitch_only
        self.preserve_rest_position = preserve_rest_position
        self.disable_rewrite_meter = disable_rewrite_meter
        self.omit_time_signatures = omit_time_signatures
        self.boundary_depth = boundary_depth
        self.maximum_dot_count = maximum_dot_count
        self.rewrite_tuplets = rewrite_tuplets
        self.prettify_rewrite_meter = prettify_rewrite_meter
        self.extract_trivial_tuplets = extract_trivial_tuplets
        self.fuse_across_groups_of_beats = fuse_across_groups_of_beats
        self.fuse_quadruple_meter = fuse_quadruple_meter
        self.fuse_triple_meter = fuse_triple_meter
        self.process_on_first_call = process_on_first_call
        self.swap_limit = swap_limit
        self._is_first_window = True

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of :attr:`contents`."""
        return abjad.lilypond(self._contents)

    def __len__(self) -> int:
        r"""Returns the number of logical ties of :attr:`contents`."""
        return len(self._logical_selections)

    def __call__(self) -> abjad.Selection:
        r"""Calls the shuffling process, returning an |abjad.Selection|"""
        return self.shuffle()

    def __next__(self) -> abjad.Selection:
        r"""Calls the shuffling process for one iteration, returning an
        |abjad.Selection|.
        """
        return self.__call__()

    def __iter__(self) -> None:
        r"""Returns an iterator, allowing instances to be used as iterators."""
        return self

    ### PUBLIC METHODS ###

    def shuffle(self) -> abjad.Selection:
        r"""Shuffles logical ties or pitches of :attr:`contents`."""
        if self._is_first_window and not self._process_on_first_call:
            if not self._pitch_only:
                self._rewrite_logical_selections()
                return self.current_window
            else:
                self._rewrite_pitches()
                return self.current_window
        else:
            if not self._pitch_only:
                return self._shuffle_logical_selections()
            else:
                return self._shuffle_pitches()

    def rotate(
        self,
        *,
        n_rotations: int = 1,
        anticlockwise: bool = False,
    ) -> abjad.Selection:
        r"""Rotates logical ties or pitches of :attr:`contents`."""
        if not isinstance(n_rotations, int):
            raise TypeError("'n_rotations' must be 'int'")
        if n_rotations < 1:
            raise ValueError("'n_rotations' must be greater than zero")
        if not isinstance(anticlockwise, bool):
            raise TypeError("'anticlockwise' must be 'bool'")
        if self._is_first_window and not self._process_on_first_call:
            if not self._pitch_only:
                self._rewrite_logical_selections()
                return self.current_window
            else:
                self._rewrite_pitches()
                return self.current_window
        else:
            if not self._pitch_only:
                return self._rotate_logical_selections(
                    n_rotations=n_rotations,
                    anticlockwise=anticlockwise,
                )
            else:
                return self._rotate_pitches(
                    n_rotations=n_rotations,
                    anticlockwise=anticlockwise,
                )

    def shuffle_n(
        self,
        n: int,
    ) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the shuffling process and outputs
        a single |abjad.Selection|.
        """
        if not isinstance(n, int):
            raise TypeError("argument must be 'int'")
        if n < 1:
            raise ValueError("argument must be greater than zero")
        dummy_container = abjad.Container()
        for _ in range(n):
            dummy_container.append(self.__call__())
        mutate.remove_repeated_time_signatures(dummy_container[:])
        mutate.remove_repeated_dynamics(dummy_container[:])
        output = dummy_container[:]
        dummy_container[:] = []
        return output

    def rotate_n(
        self,
        n: int,
        *,
        n_rotations: int = 1,
        anticlockwise: bool = False,
    ) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the pitch shuffling process and
        outputs a single |abjad.Selection|.
        """
        if not isinstance(n, int):
            raise TypeError("argument must be 'int'")
        if n < 1:
            raise ValueError("argument must be greater than zero")
        dummy_container = abjad.Container()
        for _ in range(n):
            dummy_container.append(
                self.rotate(n_rotations=n_rotations, anticlockwise=anticlockwise)
            )
        mutate.remove_repeated_time_signatures(dummy_container[:])
        mutate.remove_repeated_dynamics(dummy_container[:])
        output = dummy_container[:]
        dummy_container[:] = []
        return output

    ### PRIVATE METHODS ###

    def _update_logical_selections(self) -> None:
        r"""Updates the selection of logical ties of :attr:`contents`."""
        self._logical_selections = select.logical_selections(self._contents)
        self._logical_selections_indeces = list(range(self.__len__()))

    def _get_pitch_list(self) -> None:
        r"""Creates a :obj:`list` of all pitches in :attr:`contents`."""
        self._pitches = []
        for logical_selection in self._logical_selections:
            leaf = logical_selection.leaves()[0]
            if isinstance(leaf, abjad.Rest):
                self._pitches.append(None)
            elif isinstance(leaf, abjad.Note):
                self._pitches.append(leaf.written_pitch)
            elif isinstance(leaf, abjad.Chord):
                self._pitches.append(leaf.written_pitches)

    def _shuffle_list_preserving_rests(
        self,
        input_list: list[Any],
    ) -> None:
        r"""Shuffles a :obj:`list` while keeping rest indeces unchanged."""
        dummy_list = [input_list[i] for i in range(len(input_list)) if self._pitches[i] is not None]
        self._random_shuffle(dummy_list)
        self._replace_list_preserving_rests(dummy_list, input_list)

    def _rotate_list_preserving_rests(
        self,
        input_list: list[Any],
        *,
        n_rotations: int = 1,
        anticlockwise: bool = False,
    ) -> None:
        r"""Rotates a :obj:`list` while keeping rest indeces unchanged."""
        dummy_list = [input_list[i] for i in range(len(input_list)) if self._pitches[i] is not None]
        self._rotate_list(
            dummy_list,
            n_rotations=n_rotations,
            anticlockwise=anticlockwise,
        )
        self._replace_list_preserving_rests(dummy_list, input_list)

    def _random_shuffle(
        self,
        input_list: list[Any],
    ) -> None:
        r"""Random shuffles a :obj:`list`."""
        if self._swap_limit is None:
            random.shuffle(input_list)
        else:
            for _ in range(self._swap_limit):
                if len(input_list) > 1:
                    i, j = random.sample(range(len(input_list)), 2)
                    input_list[i], input_list[j] = input_list[j], input_list[i]

    def _replace_list_preserving_rests(
        self,
        input_list: list[Any],
        destination_list: list[Any],
    ) -> None:
        r"""Substitutes back an altered :obj:`list` while preserving rests."""
        counter = 0
        for index, pitch in enumerate(self._pitches):
            if pitch is not None:
                destination_list[index] = input_list[counter]
                counter += 1

    def _shuffle_logical_selections(self) -> abjad.Selection:
        r"""Shuffles the logical ties of :attr:`contents`."""
        if len(abjad.select(self._contents).tuplets()) > 0:
            raise ValueError(
                "'contents' contain one ore more tuplets; "
                "tuplets are currently supported only in "
                "pitch-only mode"
            )
        if not self._preserve_rest_position:
            self._random_shuffle(self._logical_selections_indeces)
        else:
            self._shuffle_list_preserving_rests(self._logical_selections_indeces)
        self._rewrite_logical_selections()
        return self.current_window

    def _shuffle_pitches(self) -> abjad.Selection:
        r"""Shuffles only the pitches of :attr:`contents`."""
        if not self._preserve_rest_position:
            self._random_shuffle(self._pitches)
        else:
            self._shuffle_list_preserving_rests(self._pitches)
        self._rewrite_pitches()
        return self.current_window

    def _rotate_logical_selections(
        self,
        *,
        n_rotations: int = 1,
        anticlockwise: bool = False,
    ) -> abjad.Selection:
        r"""Rotates the logical ties of :attr:`contents`."""
        if len(abjad.select(self._contents).tuplets()) > 0:
            raise ValueError(
                "'contents' contain one ore more tuplets; "
                "tuplets are currently supported only in "
                "pitch-only mode"
            )
        if not self._preserve_rest_position:
            self._rotate_list(
                self._logical_selections_indeces,
                n_rotations=n_rotations,
                anticlockwise=anticlockwise,
            )
        else:
            self._rotate_list_preserving_rests(
                self._logical_selections_indeces,
                n_rotations=n_rotations,
                anticlockwise=anticlockwise,
            )
        self._rewrite_logical_selections()
        return self.current_window

    def _rotate_pitches(
        self,
        *,
        n_rotations: int = 1,
        anticlockwise: bool = False,
    ) -> abjad.Selection:
        r"""Rotates the pitches of :attr:`contents`."""
        if not self._preserve_rest_position:
            self._rotate_list(
                self._pitches,
                n_rotations=n_rotations,
                anticlockwise=anticlockwise,
            )
        else:
            self._rotate_list_preserving_rests(
                self._pitches,
                n_rotations=n_rotations,
                anticlockwise=anticlockwise,
            )
        self._rewrite_pitches()
        return self.current_window

    def _rewrite_logical_selections(self) -> None:
        r"""Rewrites the logical selections of the current window."""
        # writing dummy_container in shuffled order
        dummy_container = abjad.Container()
        logical_selections = select.logical_selections(abjad.mutate.copy(self._contents))
        self._force_dynamics(logical_selections)
        for index in self._logical_selections_indeces:
            logical_selection = logical_selections[index]
            dummy_container.append(logical_selection.leaves())
        # splitting leaves at measure line points
        abjad.mutate.split(
            dummy_container[:],
            [ts.duration for ts in self._time_signatures],
            cyclic=True,
        )
        # attaching time signature structure
        mutate.enforce_time_signature(
            dummy_container,
            self._time_signatures,
            disable_rewrite_meter=True,
        )
        # handling dynamics and slurs
        mutate.reposition_dynamics(dummy_container[:])
        mutate.reposition_slurs(dummy_container[:])
        # rewrite meter
        if not self._disable_rewrite_meter:
            mutate.auto_rewrite_meter(
                dummy_container,
                meter_list=self._time_signatures,
                boundary_depth=self._boundary_depth,
                maximum_dot_count=self._maximum_dot_count,
                rewrite_tuplets=self._rewrite_tuplets,
                prettify_rewrite_meter=self._prettify_rewrite_meter,
                extract_trivial_tuplets=self._extract_trivial_tuplets,
                fuse_across_groups_of_beats=self._fuse_across_groups_of_beats,
                fuse_quadruple_meter=self._fuse_quadruple_meter,
                fuse_triple_meter=self._fuse_triple_meter,
            )
        # output
        self._is_first_window = False
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    def _rewrite_pitches(self) -> None:
        r"""Rewrites the pitches of the current window."""
        dummy_container = abjad.Container(abjad.mutate.copy(self._contents[:]))
        leaf_counter = 0
        for pitch, logical_selection in zip(
            self._pitches,
            self._logical_selections,
        ):
            logical_tie = logical_selection.leaves()
            for leaf in logical_tie:
                if pitch is None:
                    new_leaf = abjad.Rest(leaf.written_duration)
                elif isinstance(pitch, abjad.PitchSegment):
                    new_leaf = abjad.Chord(pitch, leaf.written_duration)
                    if (
                        isinstance(leaf, abjad.Rest)
                        and len(logical_tie) > 1
                        and leaf is not logical_tie[-1]
                    ):
                        abjad.attach(abjad.Tie(), new_leaf)
                else:
                    new_leaf = abjad.Note(pitch, leaf.written_duration)
                    if (
                        isinstance(leaf, abjad.Rest)
                        and len(logical_tie) > 1
                        and leaf is not logical_tie[-1]
                    ):
                        abjad.attach(abjad.Tie(), new_leaf)
                for indicator in abjad.get.indicators(leaf):
                    if isinstance(indicator, (abjad.Tie, abjad.Articulation)) and pitch is None:
                        continue
                    if isinstance(indicator, abjad.TimeSignature):
                        abjad.attach(indicator, new_leaf)
                    else:
                        abjad.attach(indicator, new_leaf)
                selection = abjad.select(dummy_container).leaf(leaf_counter)
                abjad.mutate.replace(selection, new_leaf)
                leaf_counter += 1
        # attaching time signature structure
        mutate.extract_trivial_tuplets(dummy_container[:])
        mutate.enforce_time_signature(
            dummy_container,
            self._time_signatures,
            disable_rewrite_meter=True,
        )
        # output
        self._is_first_window = False
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    def _get_lilypond_format(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return self.__repr__()

    @staticmethod
    def _remove_all_time_signatures(
        container: abjad.Container,
    ) -> None:
        r"""Removes all time signatures of an |abjad.Container|."""
        for leaf in abjad.select(container).leaves():
            if abjad.get.effective(leaf, abjad.TimeSignature):
                abjad.detach(abjad.TimeSignature, leaf)

    @staticmethod
    def _force_dynamics(
        container: abjad.Container,
    ) -> None:
        r"""Enforces the current dynamic level to every logical tie without a
        dynamic marking.
        """
        logical_ties = abjad.select(container).logical_ties()
        for logical_tie in logical_ties[1:]:
            if abjad.get.indicator(logical_tie[0], abjad.Dynamic) is None:
                index = logical_ties.index(logical_tie)
                previous_logical_tie = logical_ties[index - 1]
                if abjad.get.indicator(previous_logical_tie[0], abjad.Dynamic) is not None:
                    abjad.attach(
                        abjad.get.indicator(
                            previous_logical_tie[0],
                            abjad.Dynamic,
                        ),
                        logical_tie[0],
                    )

    @staticmethod
    def _rotate_list(
        input_list: list[Any],
        *,
        n_rotations: int = 1,
        anticlockwise: bool = False,
    ) -> None:
        r"""Rotates a :obj:`list`."""
        for _ in range(n_rotations):
            if not anticlockwise:
                element = input_list.pop(0)
                input_list.append(element)
            else:
                element = input_list.pop(-1)
                input_list.insert(0, element)

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> abjad.Container:
        r"""The |abjad.Container| to be shuffled."""
        return abjad.mutate.copy(self._contents)

    @contents.setter
    def contents(
        self,
        contents: abjad.Container,
    ) -> None:
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or child class")
        if not abjad.select(contents).leaves().are_contiguous_logical_voice():
            raise ValueError("'contents' must be contiguous logical voice")
        if isinstance(contents, abjad.Score):
            self._contents = abjad.mutate.copy(contents[0])
        elif isinstance(contents, abjad.Tuplet):
            self._contents = abjad.Container([abjad.mutate.copy(contents)])
        else:
            self._contents = abjad.mutate.copy(contents)
        dummy_container = abjad.mutate.copy(contents)
        self._current_window = dummy_container[:]
        dummy_container[:] = []
        self._update_logical_selections()
        self._get_pitch_list()
        self._time_signatures = get.time_signature_list(
            self._contents,
            do_not_use_none=True,
        )
        self._is_first_window = True

    @property
    def pitch_only(self) -> bool:
        r"""When ``True``, only the pitches will be shuffled or rotated while
        the durations remain the same.
        """
        return self._pitch_only

    @pitch_only.setter
    def pitch_only(
        self,
        pitch_only: bool,
    ) -> None:
        if not isinstance(pitch_only, bool):
            raise TypeError("'pitch_only' must be 'bool'")
        self._pitch_only = pitch_only
        # potentially new logical selections when shifting from pitch-only mode
        # to logical selections mode
        self._update_logical_selections()
        self._get_pitch_list()
        self._contents = abjad.Container(abjad.mutate.copy(self._current_window))

    @property
    def preserve_rest_position(self) -> bool:
        r"""When ``True``, shuffle operations will preserve rest positions and
        durations.
        """
        return self._preserve_rest_position

    @preserve_rest_position.setter
    def preserve_rest_position(
        self,
        preserve_rest_position: bool,
    ) -> None:
        if not isinstance(preserve_rest_position, bool):
            raise TypeError("'preserve_rest_position' must be 'bool'")
        self._preserve_rest_position = preserve_rest_position

    @property
    def disable_rewrite_meter(self) -> bool:
        r"""When ``True``, the durations of the notes in the output will not be
        rewritten by the |abjad.Meter.rewrite_meter()| mutation.
        """
        return self._disable_rewrite_meter

    @disable_rewrite_meter.setter
    def disable_rewrite_meter(
        self,
        disable_rewrite_meter: bool,
    ) -> None:
        if not isinstance(disable_rewrite_meter, bool):
            raise TypeError("'disable_rewrite_meter' must be 'bool'")
        self._disable_rewrite_meter = disable_rewrite_meter

    @property
    def omit_time_signatures(self) -> bool:
        r"""When ``True``, the output will contain no time signatures."""
        return self._omit_time_signatures

    @omit_time_signatures.setter
    def omit_time_signatures(
        self,
        omit_time_signatures: bool,
    ) -> None:
        if not isinstance(omit_time_signatures, bool):
            raise TypeError("'omit_time_signatures' must be 'bool'")
        self._omit_time_signatures = omit_time_signatures

    @property
    def boundary_depth(self) -> Union[int, None]:
        r"""Sets the argument ``boundary_depth`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._boundary_depth

    @boundary_depth.setter
    def boundary_depth(
        self,
        boundary_depth: Optional[int],
    ) -> None:
        if boundary_depth is not None:
            if not isinstance(boundary_depth, int):
                raise TypeError("'boundary_depth' must be 'int'")
        self._boundary_depth = boundary_depth

    @property
    def maximum_dot_count(self) -> Union[int, None]:
        r"""Sets the argument ``maximum_dot_count`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._maximum_dot_count

    @maximum_dot_count.setter
    def maximum_dot_count(
        self,
        maximum_dot_count: Optional[int],
    ) -> None:
        if maximum_dot_count is not None:
            if not isinstance(maximum_dot_count, int):
                raise TypeError("'maximum_dot_count' must be 'int'")
        self._maximum_dot_count = maximum_dot_count

    @property
    def rewrite_tuplets(self) -> bool:
        r"""Sets the argument ``rewrite_tuplets`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._rewrite_tuplets

    @rewrite_tuplets.setter
    def rewrite_tuplets(
        self,
        rewrite_tuplets: bool,
    ) -> None:
        if not isinstance(rewrite_tuplets, bool):
            raise TypeError("'rewrite_tuplets' must be 'bool'")
        self._rewrite_tuplets = rewrite_tuplets

    @property
    def prettify_rewrite_meter(self) -> bool:
        r"""Used to enable or disable the mutation
        |auxjad.mutate.prettify_rewrite_meter()| (default ``True``).
        """
        return self._prettify_rewrite_meter

    @prettify_rewrite_meter.setter
    def prettify_rewrite_meter(
        self,
        prettify_rewrite_meter: bool,
    ) -> None:
        if not isinstance(prettify_rewrite_meter, bool):
            raise TypeError("'prettify_rewrite_meter' must be 'bool'")
        self._prettify_rewrite_meter = prettify_rewrite_meter

    @property
    def extract_trivial_tuplets(self) -> bool:
        r"""Sets the argument ``extract_trivial_tuplets`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._extract_trivial_tuplets

    @extract_trivial_tuplets.setter
    def extract_trivial_tuplets(
        self,
        extract_trivial_tuplets: bool,
    ) -> None:
        if not isinstance(extract_trivial_tuplets, bool):
            raise TypeError("'extract_trivial_tuplets' must be 'bool'")
        self._extract_trivial_tuplets = extract_trivial_tuplets

    @property
    def fuse_across_groups_of_beats(self) -> bool:
        r"""Sets the argument ``fuse_across_groups_of_beats`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_across_groups_of_beats

    @fuse_across_groups_of_beats.setter
    def fuse_across_groups_of_beats(
        self,
        fuse_across_groups_of_beats: bool,
    ) -> None:
        if not isinstance(fuse_across_groups_of_beats, bool):
            raise TypeError("'fuse_across_groups_of_beats' must be 'bool'")
        self._fuse_across_groups_of_beats = fuse_across_groups_of_beats

    @property
    def fuse_quadruple_meter(self) -> bool:
        r"""Sets the argument ``fuse_quadruple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_quadruple_meter

    @fuse_quadruple_meter.setter
    def fuse_quadruple_meter(
        self,
        fuse_quadruple_meter: bool,
    ) -> None:
        if not isinstance(fuse_quadruple_meter, bool):
            raise TypeError("'fuse_quadruple_meter' must be 'bool'")
        self._fuse_quadruple_meter = fuse_quadruple_meter

    @property
    def fuse_triple_meter(self) -> bool:
        r"""Sets the argument ``fuse_triple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_triple_meter

    @fuse_triple_meter.setter
    def fuse_triple_meter(
        self,
        fuse_triple_meter: bool,
    ) -> None:
        if not isinstance(fuse_triple_meter, bool):
            raise TypeError("'fuse_triple_meter' must be 'bool'")
        self._fuse_triple_meter = fuse_triple_meter

    @property
    def process_on_first_call(self) -> bool:
        r"""If ``True`` then :attr:`contents` will be processed in the very
        first call.
        """
        return self._process_on_first_call

    @process_on_first_call.setter
    def process_on_first_call(
        self,
        process_on_first_call: bool,
    ) -> None:
        if not isinstance(process_on_first_call, bool):
            raise TypeError("'process_on_first_call' must be 'bool'")
        self._process_on_first_call = process_on_first_call

    @property
    def swap_limit(self) -> int:
        r"""If :obj:`int` then it dictates how instances of random swapping of
        pairs of elements are applied to :attr:`contents` in each iteration. If
        ``None`` then :attr:`contents` is completely shuffled at each
        iteration.
        """
        return self._swap_limit

    @swap_limit.setter
    def swap_limit(
        self,
        swap_limit: Optional[int],
    ) -> None:
        if swap_limit is not None:
            if not isinstance(swap_limit, int):
                raise TypeError("'swap_limit' must be 'int'")
            if swap_limit < 1:
                raise ValueError("'swap_limit' must be equal to or greater than 1")
        self._swap_limit = swap_limit

    @property
    def current_window(self) -> abjad.Selection:
        r"""Read-only property, returns the result of the last operation."""
        current_window = abjad.mutate.copy(self._current_window)
        if self._omit_time_signatures:
            self._remove_all_time_signatures(current_window)
        return current_window



import copy
from typing import Any, Union

from ._LooperParent import _LooperParent


class ListLooper(_LooperParent):
    r"""Outputs slices of a :obj:`list` using the metaphor of a looping window
    of a constant number of elements. This number is given by the argument
    :attr:`window_size`, which is an :obj:`int` representing how many elements
    are to be included in each slice.

    For instance, if the initial container had the elements
    ``[A, B, C, D, E, F]`` (where each letter represents an element of an
    arbitrary type) and the looping window was size ``3``, the output would be:

    ``[A B C] [B C D] [C D E] [D E F] [E F] [F]``

    This can be better visualised as:

    .. code-block:: none

        A B C
          B C D
            C D E
              D E F
                E F
                  F

    Basic usage:
        Calling the object will return a :obj:`list` generated by the looping
        process. Each call of the object will move the window forwards and
        output the result.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.ListLooper(input_list, window_size=3)
        >>> looper()
        ['A', 'B', 'C']
        >>> looper()
        ['B', 'C', 'D']

        The property :attr:`current_window` can be used to access the current
        window without moving the head forwards.

        >>> looper.current_window
        ['B', 'C', 'D']

    :attr:`process_on_first_call`:
        The very first call will output the input :obj:`list` without
        processing it. To disable this behaviour and have the looping window
        move on the very first call, initialise the class with the keyword
        argument :attr:`process_on_first_call` set to ``True``.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.ListLooper(input_list,
        ...                            window_size=3,
        ...                            process_on_first_call=True,
        ...                            )
        >>> looper()
        ['B', 'C', 'D']

    Using as iterator:
        The instances of this class can also be used as an iterator, which can
        then be used in a for loop to exhaust all windows.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.ListLooper(input_list,
        ...                            window_size=3,
        ...                            )
        >>> for window in looper:
        ...     print(window)
        ['A', 'B', 'C']
        ['B', 'C', 'D']
        ['C', 'D', 'E']
        ['D', 'E', 'F']
        ['E', 'F']
        ['F']

    Arguments and properties:
        This class can take many optional keyword arguments during its
        creation. :attr:`step_size` dictates the size of each individual step
        in number of elements (default value is ``1``). :attr:`max_steps` sets
        the maximum number of steps that the window can advance when the object
        is called, ranging between ``1`` and the input value (default is also
        ``1``). :attr:`repetition_chance` sets the chance of a window result
        repeating itself (that is, the window not moving forwards when called).
        It should range from ``0.0`` to ``1.0`` (default ``0.0``, i.e. no
        repetition). :attr:`forward_bias` sets the chance of the window moving
        forward instead of backwards. It should range from ``0.0`` to ``1.0``
        (default ``1.0``, which means the window can only move forwards. A
        value of ``0.5`` gives 50% chance of moving forwards while a value of
        ``0.0`` will move the window only backwards). :attr:`head_position` can
        be used to offset the starting position of the looping window. It must
        be an :obj:`int` and its default value is ``0``. By default, calling
        the object will first return the original container and subsequent
        calls will process it; set :attr:`process_on_first_call` to ``True``
        and the looping process will be applied on the very first call. Lastly,
        set :attr:`end_with_max_n_elements` to ``True`` to end the process when
        the final window has the maximum number of elements.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.ListLooper(input_list,
        ...                            window_size=3,
        ...                            step_size=1,
        ...                            max_steps=2,
        ...                            repetition_chance=0.25,
        ...                            forward_bias=0.2,
        ...                            head_position=0,
        ...                            end_with_max_n_elements=True,
        ...                            process_on_first_call=True,
        ...                            )
        >>> looper.window_size
        3
        >>> looper.step_size
        1
        >>> looper.repetition_chance
        0.25
        >>> looper.forward_bias
        0.2
        >>> looper.max_steps
        2
        >>> looper.head_position
        0
        >>> looper.end_with_max_n_elements
        True
        >>> looper.process_on_first_call
        True

        Use the properties below to change these values after initialisation.

        >>> looper.window_size = 2
        >>> looper.step_size = 2
        >>> looper.max_steps = 3
        >>> looper.repetition_chance = 0.1
        >>> looper.forward_bias = 0.8
        >>> looper.head_position = 2
        >>> looper.end_with_max_n_elements = False
        >>> looper.process_on_first_call = False
        >>> looper.window_size
        2
        >>> looper.step_size
        2
        >>> looper.max_steps
        3
        >>> looper.repetition_chance
        0.1
        >>> looper.forward_bias
        0.8
        >>> looper.head_position
        2
        >>> looper.end_with_max_n_elements
        False
        >>> looper.process_on_first_call
        False

    Setting :attr:`forward_bias` to ``0.0``:
        Set :attr:`forward_bias` to ``0.0`` to move backwards instead of
        forwards (default is ``1.0``). The initial :attr:`head_position` must
        be greater than ``0`` otherwise the contents will already be exhausted
        in the very first call (since it will not be able to move backwards
        from that position).

        >>> input_list = ['A', 'B', 'C', 'D']
        >>> looper = auxjad.ListLooper(input_list,
        ...                            window_size=2,
        ...                            head_position=2,
        ...                            forward_bias=0.0,
        ...                            )
        >>> looper.output_all()
        ['C', 'D', 'B', 'C', 'A', 'B']

    :attr:`forward_bias` between ``0.0`` and ``1.0``:
        Setingt :attr:`forward_bias` to a value in between ``0.0`` and ``1.0``
        will result in random steps being taken forward or backward, according
        to the bias. The initial value of :attr:`head_position` will once gain
        play an important role here, as the contents might be exhausted if the
        looper attempts to move backwards after reaching the head position
        ``0``.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H']
        >>> looper = auxjad.ListLooper(input_list,
        ...                            window_size=2,
        ...                            head_position=4,
        ...                            forward_bias=0.5,
        ...                            )
        >>> looper.output_n(4)
        ['E', 'F', 'D', 'E', 'C', 'D', 'B', 'C']

    :attr:`max_steps`:
        Setting the keyword argument :attr:`max_steps` to a value larger than
        ``1`` will result in a random number of steps (between ``1`` and
        :attr:`max_steps`) being applied at each call.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H']
        >>> looper = auxjad.ListLooper(input_list,
        ...                            window_size=2,
        ...                            head_position=2,
        ...                            max_steps=4,
        ...                            )
        >>> looper.output_n(4)
        ['C', 'D', 'D', 'E', 'E', 'F', 'H']

    :func:`len()`:
        The function :func:`len()` can be used to get the total number of
        elements in the container.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.ListLooper(input_list, window_size=3)
        >>> len(looper)
        6

    :meth:`output_all`:
        To run through the whole process and output it as a single :obj:`list`,
        from the initial head position until the process outputs the single
        last element, use the method :meth:`output_all`.

        >>> input_list = ['A', 'B', 'C', 'D']
        >>> looper = auxjad.ListLooper(input_list, window_size=3)
        >>> looper.output_all()
        ['A', 'B', 'C', 'B', 'C', 'D', 'C', 'D', 'D']

    :meth:`output_n`:
        To run through just part of the process and output it as a single
        :obj:`list`, starting from the initial head position, use the method
        :meth:`output_n` and pass the number of iterations as argument.

        >>> input_list = ['A', 'B', 'C', 'D']
        >>> looper = auxjad.ListLooper(input_list, window_size=3)
        >>> looper.output_n(2)
        ['A', 'B', 'C', 'B', 'C', 'D']

    :attr:`end_with_max_n_elements`:
        When ``True``, the last bar in the output will contain the maximum
        number of leaves given by :attr:`window_size`. E.g. consider the
        elements ``[A, B, C, D]`` and the looping window was size ``3``;
        setting :attr:`end_with_max_n_elements` to ``True`` will output:

        ``[A B C] [B C D]``

        Setting it to ``False`` (which is this property's default value) will
        produces:

        ``[A B C] [B C D] [C D] [D]``

        Compare the two examples below:

        >>> input_list = ['A', 'B', 'C', 'D']
        >>> looper = auxjad.ListLooper(input_list, window_size=3)
        >>> looper.output_all()
        ['A', 'B', 'C', 'B', 'C', 'D', 'C', 'D', 'D']
        >>> input_list = ['A', 'B', 'C', 'D']
        >>> looper = auxjad.ListLooper(input_list,
        ...                            window_size=3,
        ...                            end_with_max_n_elements=True,
        ...                            )
        >>> looper.output_all()
        ['A', 'B', 'C', 'B', 'C', 'D']

    :attr:`window_size`:
        To change the size of the looping window after instantiation, use the
        property :attr:`window_size`. In the example below, the initial window
        is of size ``3``, and so the first call of the looper object outputs
        the first, second, and third elements of the :obj:`list`. The window
        size is then set to ``4``, and the looper is called again, moving to
        the element in the next position, thus outputting the second, third,
        fourth, and fifth elements.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.ListLooper(input_list, window_size=3)
        >>> looper()
        ['A', 'B', 'C']
        >>> looper.window_size = 4
        >>> looper()
        ['B', 'C', 'D', 'E']

    :attr:`contents`:
        Use the :attr:`contents` property to read as well as overwrite the
        contents of the looper. Notice that the :attr:`head_position` will
        remain on its previous value and must be reset to ``0`` if that's
        required.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.ListLooper(input_list,
        ...                            window_size=3,
        ...                            )
        >>> looper.contents
        ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper()
        ['A', 'B', 'C']
        >>> looper()
        ['B', 'C', 'D']
        >>> looper.contents = [0, 1, 2, 3, 4]
        >>> looper.contents
        [0, 1, 2, 3, 4]
        >>> looper()
        [1, 2, 3]
        >>> looper.head_position = 0
        >>> looper()
        [0, 1, 2]

    Types in the input list:
        The input :obj:`list` can contain any types of elements:

        >>> input_list = [123, 'foo', (3, 4), 3.14]
        >>> looper = auxjad.ListLooper(input_list, window_size=3)
        >>> looper()
        [123, 'foo', (3, 4)]

        This also include Abjad's types. Abjad's exclusive membership
        requirement is respected since each call returns a
        :func:`copy.deepcopy` of the window. The same is true to the
        :meth:`output_all` method.

        >>> import abjad
        >>> import copy
        >>> input_list = [
        ...     abjad.Container(r"c'4 d'4 e'4 f'4"),
        ...     abjad.Container(r"fs'1"),
        ...     abjad.Container(r"r2 bf4 c'4"),
        ...     abjad.Container(r"c''2. r4"),
        ... ]
        >>> looper = auxjad.ListLooper(input_list, window_size=3)
        >>> staff = abjad.Staff()
        >>> for element in looper.output_all():
        ...     staff.append(element)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }
                {
                    fs'1
                }
                {
                    r2
                    bf4
                    c'4
                }
                {
                    fs'1
                }
                {
                    r2
                    bf4
                    c'4
                }
                {
                    c''2.
                    r4
                }
                {
                    r2
                    bf4
                    c'4
                }
                {
                    c''2.
                    r4
                }
                {
                    c''2.
                    r4
                }
            }

        ..  figure:: ../_images/ListLooper-kvxaoz53y5f.png
    """

    ### CLASS VARIABLES ###

    __slots__ = "_end_with_max_n_elements"

    ### INITIALISER ###

    def __init__(
        self,
        contents: list[Any],
        *,
        window_size: int,
        step_size: int = 1,
        max_steps: int = 1,
        repetition_chance: float = 0.0,
        forward_bias: float = 1.0,
        head_position: int = 0,
        end_with_max_n_elements: bool = False,
        process_on_first_call: bool = False,
    ) -> None:
        self.contents = contents
        self.end_with_max_n_elements = end_with_max_n_elements
        super().__init__(
            head_position=head_position,
            window_size=window_size,
            step_size=step_size,
            max_steps=max_steps,
            repetition_chance=repetition_chance,
            forward_bias=forward_bias,
            process_on_first_call=process_on_first_call,
        )

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of :attr:`contents`."""
        return str(self._contents)

    def __len__(self) -> int:
        r"""Returns a length of :attr:`contents`."""
        return len(self._contents)

    ### PUBLIC METHODS ###

    def output_all(self) -> list[Any]:
        r"""Goes through the whole looping process and outputs a single
        :obj:`list`. This method replaces the parent's one since the parent's
        method outputs an |abjad.Selection|.
        """
        dummy_container = []
        while True:
            try:
                dummy_container.extend(self.__call__())
            except RuntimeError:
                break
        return dummy_container[:]

    def output_n(self, n: int) -> list[Any]:
        r"""Goes through ``n`` iterations of the looping process and outputs a
        single :obj:`list`. This method replaces the parent's one since the
        parent's method outputs an |abjad.Selection|.
        """
        if not isinstance(n, int):
            raise TypeError("argument must be 'int'")
        if n < 0:
            raise ValueError("argument must be a positive 'int'")
        dummy_container = []
        for _ in range(n):
            dummy_container.extend(self.__call__())
        return dummy_container[:]

    ### PRIVATE METHODS ###

    def _slice_contents(self) -> None:
        r"""This method takes a slice with :attr:`window_size` number of
        elements out of :attr:`contents` starting at the current
        :attr:`head_position`.
        """
        start = self._head_position
        end = self._head_position + self._window_size
        self._current_window = self._contents[start:end]

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> list[Any]:
        r"""The :obj:`list` to be sliced and looped."""
        return self._contents

    @contents.setter
    def contents(
        self,
        contents: list[Any],
    ) -> None:
        if not isinstance(contents, list):
            raise TypeError("'contents' must be 'list")
        self._contents = contents[:]
        self._is_first_window = True

    @property
    def current_window(self) -> Union[list[Any], None]:
        r"""Read-only property, returns the window at the current head
        position.
        """
        if self._current_window is None:
            return self._current_window
        return copy.deepcopy(self._current_window)[:]

    @property
    def end_with_max_n_elements(self) -> bool:
        r"""When ``True``, the last bar in the output will contain the maximum
        number of elements given by :attr:`window_size`. E.g. consider the
        elements ``[A, B, C, D]`` and the looping window was size ``3``;
        setting :attr:`end_with_max_n_elements` to ``True`` will output:

        ``[A B C] [B C D]``

        Setting it to ``False`` (which is this property's default value) will
        produces:

        ``[A B C] [B C D] [C D] [D]``
        """
        return self._end_with_max_n_elements

    @end_with_max_n_elements.setter
    def end_with_max_n_elements(
        self,
        end_with_max_n_elements: bool,
    ) -> None:
        if not isinstance(end_with_max_n_elements, bool):
            raise TypeError("'end_with_max_n_elements' must be 'bool'")
        self._end_with_max_n_elements = end_with_max_n_elements

    ### PRIVATE PROPERTIES ###

    @property
    def _done(self) -> bool:
        r""":obj:`bool` indicating whether the process is done (i.e. whether
        the head position has overtaken the :attr:`contents`'s length).
        """
        if self._end_with_max_n_elements:
            return (
                self._head_position + self._window_size > self.__len__() or self._head_position < 0
            )
        else:
            return self._head_position >= self.__len__() or self._head_position < 0



import abjad


class TimeSignature(abjad.TimeSignature):
    r"""Time signature."""

    def simplify_ratio(
        self,
        *,
        min_denominator: int = 4,
    ) -> None:
        r"""Simplifies the ratio of an |abjad.TimeSignature| according to a
        minimum denominator value. Mutates the ratio in place and has no return
        value. Note that Auxjad adds this function as an extension method to
        |abjad.TimeSignature| (see usage below).

        Basic usage:
            By default, this method simplifies the ratio of
            ``numerator / denominator`` using a minimum denominator value of
            ``4`` (that is, the denominator will not get smaller than ``4``).
            In the case below, ``(2, 4)`` is the simplest representation of the
            ratio ``(4, 8)`` with a denominator equal to or larger than ``4``.

            >>> time_signature = abjad.TimeSignature((4, 8))
            >>> time_signature.simplify_ratio()
            >>> time_signature
            abjad.TimeSignature((2, 4))
            >>> time_signature = abjad.TimeSignature((1, 1))
            >>> time_signature.simplify_ratio()
            >>> time_signature
            abjad.TimeSignature((4, 4))

            If a ratio cannot be simplified at all, the function returns a time
            signature with the original ratio.

            >>> time_signature = abjad.TimeSignature((7, 8))
            >>> time_signature.simplify_ratio()
            >>> time_signature
            abjad.TimeSignature((7, 8))

            It also simplifies ratios even if the minimum denominator cannot be
            reached.

            >>> time_signature = abjad.TimeSignature((10, 16))
            >>> time_signature.simplify_ratio()
            >>> time_signature
            abjad.TimeSignature((5, 8))

        ``min_denominator``:
            The ``min_denominator`` can be set to values other than ``4``. If
            set to ``2``, the simplest representaion of the ratio ``(4, 8)``
            becomes ``(1, 2)``.

            >>> time_signature = abjad.TimeSignature((4, 8))
            >>> time_signature.simplify_ratio(min_denominator=2)
            >>> time_signature
            abjad.TimeSignature((1, 2))
            >>> time_signature = abjad.TimeSignature((1, 1))
            >>> time_signature.simplify_ratio(min_denominator=1)
            >>> time_signature
            abjad.TimeSignature((1, 1))
        """
        if not isinstance(min_denominator, int):
            raise TypeError("'min_denominator' must be 'int'")

        while self._denominator < min_denominator:
            self._numerator *= 2
            self._denominator *= 2
        while True:
            if (
                self._numerator % 2 == 0
                and self._denominator % 2 == 0
                and self._denominator >= 2 * min_denominator
            ):
                self._numerator //= 2
                self._denominator //= 2
            else:
                return


### EXTENSION METHODS ###

abjad.TimeSignature.simplify_ratio = TimeSignature.simplify_ratio



import abjad

from .. import get


def close_container(container: abjad.Container) -> None:
    r"""Mutates an input container (of type |abjad.Container| or child class)
    in place and has no return value; this function changes the time signature
    of the last measure of an underfull in order to make it full.

    Basic usage:
        Returns the missing duration of the last measure of any container or
        child class. If no time signature is encountered, it uses LilyPond's
        convention and considers the container as in 4/4.

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4")
        >>> container3 = abjad.Staff(r"c'4 d'4 e'4 f'4 | c'4")
        >>> container4 = abjad.Staff(r"c'4 d'4 e'4 f'4 | c'4 d'4 e'4 f'4")
        >>> auxjad.mutate.close_container(container1)
        >>> auxjad.mutate.close_container(container2)
        >>> auxjad.mutate.close_container(container3)
        >>> auxjad.mutate.close_container(container4)
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/close_container-nfnk06s90x.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/close_container-swyxwiup8pm.png

        >>> abjad.show(container3)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
                \time 1/4
                c'4
            }

        ..  figure:: ../_images/close_container-mms5hiysbwe.png

        >>> abjad.show(container4)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/close_container-aky7avbla4w.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.close_containers(staff)
        >>> abjad.mutate.close_containers(staff)

    Time signature changes:
        Handles any time signatures as well as changes of time signature.

        >>> container1 = abjad.Staff(r"\time 4/4 c'4 d'4 e'4 f'4 g'")
        >>> container2 = abjad.Staff(r"\time 3/4 a2. \time 2/4 c'4")
        >>> container3 = abjad.Staff(r"\time 5/4 g1 ~ g4 \time 4/4 af'2")
        >>> auxjad.mutate.close_container(container1)
        >>> auxjad.mutate.close_container(container2)
        >>> auxjad.mutate.close_container(container3)
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
                \time 1/4
                g'4
            }

        ..  figure:: ../_images/close_container-3tgyty245cq.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 3/4
                a2.
                \time 1/4
                c'4
            }

        ..  figure:: ../_images/close_container-st5d89zofoh.png

        >>> abjad.show(container3)

        ..  docs::

            \new Staff
            {
                \time 5/4
                g1
                ~
                g4
                \time 2/4
                af'2
            }

        ..  figure:: ../_images/close_container-wd5irlm76l.png

    ..  note::

        When using |abjad.Container|'s, all time signatures in the output will
        be commented out with ``%%%.`` This is because Abjad only applies time
        signatures to containers that belong to a |abjad.Staff|. The present
        function works with either |abjad.Container| and |abjad.Staff|.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> abjad.show(container)

        ..  docs::

            {
                %%% \time 3/4 %%%
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/close_container-xrwkorhtl5.png

        >>> staff = abjad.Staff([container])
        >>> abjad.show(container)

        ..  docs::

            {
                \time 3/4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/close_container-ys6pcdlywy.png

    Partial time signatures:
        Correctly handles partial time signatures.

        >>> container = abjad.Staff(r"c'4 d'4 e'4 f'4 g'4")
        >>> time_signature = abjad.TimeSignature((3, 4), partial=(1, 4))
        >>> abjad.attach(time_signature, container[0])
        >>> auxjad.mutate.close_container(container)
        >>> abjad.show(container)

        ..  docs::

            \new Staff
            {
                \partial 4
                \time 3/4
                c'4
                d'4
                e'4
                f'4
                \time 1/4
                g'4
            }

        ..  figure:: ../_images/close_container-rfskjbbfnu.png

    ..  error::

        If a container is malformed, i.e. it has an underfilled measure before
        a time signature change, the function raises a :exc:`ValueError`
        exception.

        >>> container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'4")
        >>> auxjad.mutate.close_container(container)
        ValueError: 'container' is malformed, with an underfull measure
        preceding a time signature change

    ..  warning::

        The input container must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("argument must be 'abjad.Container' or child class")
    if not abjad.select(container).leaves().are_contiguous_logical_voice():
        raise ValueError("argument must be contiguous logical voice")
    try:
        if not get.selection_is_full(container[:]):
            missing_duration = get.underfull_duration(container[:])
            leaves = abjad.select(container).leaves()
            for leaf in leaves[::-1]:
                time_signature = abjad.get.effective(leaf, abjad.TimeSignature)
                if time_signature is not None:
                    last_time_signature = time_signature
                    break
            else:
                last_time_signature = abjad.TimeSignature((4, 4))
            last_bar_duration = last_time_signature.duration - missing_duration
            final_bar_time_signature = abjad.TimeSignature(last_bar_duration)
            final_bar_time_signature.simplify_ratio()
            duration = 0
            for leaf in leaves[::-1]:
                duration += abjad.get.duration(leaf)
                if duration == last_bar_duration:
                    if abjad.get.indicator(leaf, abjad.TimeSignature):
                        abjad.detach(abjad.TimeSignature, leaf)
                    abjad.attach(final_bar_time_signature, leaf)
                    break
    except ValueError as err:
        raise ValueError(
            "'container' is malformed, with an underfull measure "
            "preceding a time signature change"
        ) from err



"""
Auxjad
======

Auxjad is a library of auxiliary classes and functions for Abjad 3.4 aimed at
composers of algorithmic music. All classes and functions have a ``__doc__``
    attribute with usage instructions.

Documentation is available at https://gilbertohasnofb.github.io/auxjad-docs/.

Bugs can be reported through the project's Issue Tracker.

This library is published under the MIT License.
"""

from . import get, mutate, select

from .core.CartographySelector import CartographySelector
from .core.CrossFader import CrossFader
from .core.Echoer import Echoer
from .core.Fader import Fader
from .core.GeneticAlgorithm import GeneticAlgorithm
from .core.Hocketer import Hocketer
from .core.LeafLooper import LeafLooper
from .core.ListLooper import ListLooper
from .core.Phaser import Phaser
from .core.PitchRandomiser import PitchRandomiser
from .core.Repeater import Repeater
from .core.Shuffler import Shuffler
from .core.TenneySelector import TenneySelector
from .core.WindowLooper import WindowLooper

from .indicators.TimeSignature import TimeSignature

from .makers.GeneticAlgorithmMusicMaker import GeneticAlgorithmMusicMaker
from .makers.LeafDynMaker import LeafDynMaker

from .score.ArtificialHarmonic import ArtificialHarmonic
from .score.HarmonicNote import HarmonicNote
from .score.Score import Score

from .spanners.piano_pedal import piano_pedal

from .utilities.staff_splitter import staff_splitter

from ._version import __version__

__author__ = "Gilberto Agostinho <gilbertohasnofb@gmail.com>"
__all__ = [
    "__author__",
    "__version__",
    # modules
    "get",
    "mutate",
    "select",
    # core
    "CartographySelector",
    "CrossFader",
    "Echoer",
    "Fader",
    "GeneticAlgorithm",
    "Hocketer",
    "LeafLooper",
    "ListLooper",
    "Phaser",
    "PitchRandomiser",
    "Repeater",
    "Shuffler",
    "TenneySelector",
    "WindowLooper",
    # indicators
    "TimeSignature",
    # makers
    "GeneticAlgorithmMusicMaker",
    "LeafDynMaker",
    # score
    "ArtificialHarmonic",
    "HarmonicNote",
    "Score",
    # spanners
    "piano_pedal",
    # utilities
    "staff_splitter",
]



import abjad


def merge_hairpins(selection: abjad.Selection) -> None:
    r"""Mutates an input |abjad.Selection| in place and has no return value;
    this function merges all consecutive hairpins that share a same trend.

    Basic usage:
        Merging crescendo hairpins:

        >>> staff = abjad.Staff(r"c'4\pp\< d'4 e'4\p\< f'4 g'1\mp")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                \<
                d'4
                e'4
                \p
                \<
                f'4
                g'1
                \mp
            }

        ..  figure:: ../_images/merge_hairpins-b4it6YReZ9.png

        >>> abjad.mutate.merge_hairpins(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                \<
                d'4
                e'4
                f'4
                g'1
                \mp
            }

        ..  figure:: ../_images/merge_hairpins-s5cU8PINVO.png

        Merging diminuendo hairpins:

        >>> staff = abjad.Staff(r"c'4\ff\> d'4 e'4\mf\> f'4 g'1\pp")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \ff
                \>
                d'4
                e'4
                \mf
                \>
                f'4
                g'1
                \pp
            }

        ..  figure:: ../_images/merge_hairpins-GUIqXKDXaQ.png

        >>> abjad.mutate.merge_hairpins(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \ff
                \>
                d'4
                e'4
                f'4
                g'1
                \pp
            }

        ..  figure:: ../_images/merge_hairpins-adbT777qMg.png

        The merge only occurs when hairpins share a trend:

        >>> staff = abjad.Staff(r"c'4\pp\< d'4 e'4\p\> f'4 g'1\pp")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                \<
                d'4
                e'4
                \p
                \>
                f'4
                g'1
                \pp
            }

        ..  figure:: ../_images/merge_hairpins-Lvzn2f6PU6.png

        >>> abjad.mutate.merge_hairpins(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                \<
                d'4
                e'4
                \p
                \>
                f'4
                g'1
                \pp
            }

        ..  figure:: ../_images/merge_hairpins-EJdWJwkUJE.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate(staff[:]).merge_hairpins()
        >>> abjad.mutate(staff[:]).merge_hairpins()

    Gaps:
        If there are any gaps between hairpins (i.e. they finish on an earlier
        leaf than the start of the new hairpin), they will not be merged even
        if they share a trend.

        >>> staff = abjad.Staff(
        ...     r"c'1\pp\< d'2\f d'2\< e'1\ff f'1\f\> "
        ...     r"g'2\mp g'2\> a'1\! b'1\ppp"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \pp
                \<
                d'2
                \f
                d'2
                \<
                e'1
                \ff
                f'1
                \f
                \>
                g'2
                \mp
                g'2
                \>
                a'1
                \!
                b'1
                \ppp
            }

        ..  figure:: ../_images/merge_hairpins-3Bo4GAfHaX.png

        >>> auxjad.mutate.merge_hairpins(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \pp
                \<
                d'2
                \f
                d'2
                \<
                e'1
                \ff
                f'1
                \f
                \>
                g'2
                \mp
                g'2
                \>
                a'1
                \!
                b'1
                \ppp
            }

        ..  figure:: ../_images/merge_hairpins-HaF1ZUdqtx.png

    Multiple merges:
        This function can merge an indefinite number of hairpins:

        >>> staff = abjad.Staff(
        ...     r"c'4\pp\< d'4 e'4\p\< f'4 g'4\mp\< a'4 b'4\mf\< c''4"
        ...     r"d''4\f\> c''4 b'4\mf\> a'4 g'4\mp\> f'4 e'4\p\> d'4"
        ...     r"c'1\pp"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                \<
                d'4
                e'4
                \p
                \<
                f'4
                g'4
                \mp
                \<
                a'4
                b'4
                \mf
                \<
                c''4
                d''4
                \f
                \>
                c''4
                b'4
                \mf
                \>
                a'4
                g'4
                \mp
                \>
                f'4
                e'4
                \p
                \>
                d'4
                c'1
                \pp
            }

        ..  figure:: ../_images/merge_hairpins-DrT5XyaGWQ.png

        >>> auxjad.mutate.merge_hairpins(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \pp
                \<
                d'4
                e'4
                f'4
                g'4
                a'4
                b'4
                c''4
                d''4
                \f
                \>
                c''4
                b'4
                a'4
                g'4
                f'4
                e'4
                d'4
                c'1
                \pp
            }

        ..  figure:: ../_images/merge_hairpins-Z2yDxl8dlI.png
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Selection'")
    active_hairpin = None
    for leaf in selection.leaves():
        leaf_hairpin = abjad.get.indicator(leaf, abjad.StartHairpin)
        if active_hairpin is not None and leaf_hairpin is not None:
            if active_hairpin.shape == leaf_hairpin.shape:
                abjad.detach(abjad.StartHairpin, leaf)
                if abjad.get.indicator(leaf, abjad.Dynamic) is not None:
                    abjad.detach(abjad.Dynamic, leaf)
        if (
            abjad.get.indicator(leaf, abjad.Dynamic) is not None
            or abjad.get.indicator(leaf, abjad.StopHairpin) is not None
        ):
            if leaf_hairpin is None:
                active_hairpin = None
        if leaf_hairpin is not None:
            active_hairpin = leaf_hairpin



from typing import Union

import abjad


class _HarmonicParent:
    r"""This is the parent class of :class:`auxjad.ArtificialHarmonic` and
    :class:`auxjad.HarmonicNote`. It implements two common properties with
    ``@property`` and ``@setter`` decorators.
    """

    ### CLASS VARIABLES ###

    # due to subclasses using multiple inheritance, '_markup' and '_direction'
    # are added to their __slots__ since you cannot inherit from multiple
    # classes with non-empty __slots__
    __slots__ = ()

    ### PUBLIC PROPERTIES ###

    @property
    def direction(self) -> Union[str, abjad.enums.VerticalAlignment]:
        r"""The direction of the harmonic note head."""
        return self._direction

    @direction.setter
    def direction(
        self,
        direction: Union[str, abjad.enums.VerticalAlignment],
    ) -> None:
        if direction is not None:
            if not isinstance(direction, (str, abjad.enums.VerticalAlignment)):
                raise TypeError(
                    "'direction' must be 'str', None, or either 'abjad.Up' or 'abjad.Down'"
                )
        self._direction = direction
        # detaching and reattaching markup to apply new direction
        markup = self._markup
        self.markup = None
        self.markup = markup

    @property
    def markup(self) -> str:
        r"""The markup of the harmonic note head."""
        return self._markup

    @markup.setter
    def markup(
        self,
        markup: str,
    ) -> None:
        if markup is not None:
            if not isinstance(markup, str):
                raise TypeError("'markup' must be 'str'")
            self._markup = markup
            markup = abjad.Markup(
                self._markup,
                direction=self._direction,
            )
            abjad.attach(markup, self)
        else:
            self._markup = markup
            if abjad.get.indicator(self, abjad.Markup):
                abjad.detach(abjad.Markup, self)



from typing import Optional, Union

import abjad

from .. import mutate
from ._LooperParent import _LooperParent


class LeafLooper(_LooperParent):
    r"""Outputs slices of an |abjad.Container| (or child class) using the
    metaphor of a looping window of a constant number of elements. This number
    is given by the argument :attr:`window_size`, which is an :obj:`int`
    representing how many notes are to be included in each slice. The duration
    of the slice will be the sum of the duration of these notes.

    For instance, if the initial container had the logical ties
    :math:`[A, B, C, D, E, F]` (where each letter represents one logical tie)
    and the looping window was size ``3``, the output would be:

    :math:`A~ B~ C~ |~ B~ C~ D~ |~ C~ D~ E~ |~ D~ E~ F~ |~ E~ F~ |~ F`

    This can be better visualised as:

    .. code-block:: none

        A B C
          B C D
            C D E
              D E F
                E F
                  F

    Basic usage:
        Calling the object will return an |abjad.Selection| generated by the
        looping process. It takes a container (or child class equivalent) and
        the number of elements of the window as arguments. Each call of the
        object will move the window forwards and output the result.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> abjad.show(container)

        ..  docs::

            {
                c'4
                d'2
                e'4
                f'2
                ~
                f'8
                g'4.
            }

        ..  figure:: ../_images/LeafLooper-IKft5wLQ8J.png

        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'2
                e'4
            }

        ..  figure:: ../_images/LeafLooper-o372nj7r3sb.png

        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 11/8
                d'2
                e'8
                ~
                e'8
                f'8
                ~
                f'2
            }

        ..  figure:: ../_images/LeafLooper-lszpb96x0rf.png

        The property :attr:`current_window` can be used to access the current
        window without moving the head forwards.

        >>> notes = looper.current_window
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 11/8
                d'2
                e'8
                ~
                e'8
                f'8
                ~
                f'2
            }

        ..  figure:: ../_images/LeafLooper-2c1jo6ftfnu.png

    :attr:`process_on_first_call`:
        The very first call will output the input container without processing
        it. To disable this behaviour and have the looping window move on the
        very first call, initialise the class with the keyword argument
        :attr:`process_on_first_call` set to ``True``.

        >>> container = abjad.Container(r"c'4 d'8 e'2 f'8 g'4")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            process_on_first_call=True,
        ...                            )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                d'8
                e'8
                ~
                e'4.
                f'8
            }

        ..  figure:: ../_images/LeafLooper-uvu1p1g04ne.png

    Using as iterator:
        The instances of this class can also be used as an iterator, which can
        then be used in a for loop to exhaust all windows. Note that unlike the
        methods :meth:`output_n` and :meth:`output_all`, time signatures are
        added to each window returned by the shuffler. Use the function
        |auxjad.mutate.remove_repeated_time_signatures()| to clean the output
        when using this class in this way.

        >>> container = abjad.Container(r"c'4 d'2 e'8 f'2")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=2,
        ...                            )
        >>> staff = abjad.Staff()
        >>> for window in looper:
        ...     staff.append(window)
        >>> auxjad.mutate.remove_repeated_time_signatures(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'2
                \time 5/8
                d'2
                e'8
                e'8
                f'2
                \time 2/4
                f'2
            }

        ..  figure:: ../_images/LeafLooper-tlddw14wnrm.png

    Arguments and properties:
        This class can take many optional keyword arguments during its
        creation. :attr:`step_size` dictates the size of each individual step
        in number of elements (default value is ``1``). :attr:`max_steps` sets
        the maximum number of steps that the window can advance when the object
        is called, ranging between ``1`` and the input value (default is also
        ``1``). :attr:`repetition_chance` sets the chance of a window result
        repeating itself (that is, the window not moving forwards when called).
        It should range from ``0.0`` to ``1.0`` (default ``0.0``, i.e. no
        repetition). :attr:`forward_bias` sets the chance of the window moving
        forward instead of backwards. It should range from ``0.0`` to ``1.0``
        (default ``1.0``, which means the window can only move forwards. A
        value of ``0.5`` gives :math:`50\%` chance of moving forwards while a
        value of ``0.0`` will move the window only backwards).
        :attr:`head_position` can be used to offset the starting position of
        the looping window. It must be an :obj:`int` and its default value is
        ``0``. Set :attr:`end_with_max_n_leaves` to ``True`` to end the process
        when the final window has the maximum number of leaves. The attribute
        :attr:`after_rest` sets the length of rests that separate consecutive
        iterations (default is ``abjad.Duration(0)``, i.e. no rest). When using
        after rests, :attr:`after_rest_in_new_measure` dictates whether or not
        rests are appended to the same output measure or to a new one. When in
        their own measure, :attr:`use_multimeasure_rests` sets if multi-measure
        rests should be used or not.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            step_size=1,
        ...                            max_steps=2,
        ...                            repetition_chance=0.25,
        ...                            forward_bias=0.2,
        ...                            head_position=0,
        ...                            end_with_max_n_leaves=True,
        ...                            omit_time_signatures=False,
        ...                            process_on_first_call=True,
        ...                            after_rest=(1, 8),
        ...                            after_rest_in_new_measure=True,
        ...                            use_multimeasure_rests=False,
        ...                            )
        >>> looper.window_size
        3
        >>> looper.step_size
        1
        >>> looper.repetition_chance
        0.25
        >>> looper.forward_bias
        0.2
        >>> looper.max_steps
        2
        >>> looper.head_position
        0
        >>> looper.end_with_max_n_leaves
        True
        >>> looper.omit_time_signatures
        False
        >>> looper.process_on_first_call
        True
        >>> looper.after_rest
        abjad.Duration((1, 8))
        >>> looper.after_rest_in_new_measure
        True
        >>> looper.use_multimeasure_rests
        False

        Use the properties below to change these values after initialisation.

        >>> looper.window_size = 2
        >>> looper.step_size = 2
        >>> looper.max_steps = 3
        >>> looper.repetition_chance = 0.1
        >>> looper.forward_bias = 0.8
        >>> looper.head_position = 2
        >>> looper.end_with_max_n_leaves = False
        >>> looper.omit_time_signatures = True
        >>> looper.process_on_first_call = False
        >>> looper.after_rest = 0
        >>> looper.after_rest_in_new_measure = False
        >>> looper.use_multimeasure_rests = True
        >>> looper.window_size
        2
        >>> looper.step_size
        2
        >>> looper.max_steps
        3
        >>> looper.repetition_chance
        0.1
        >>> looper.forward_bias
        0.8
        >>> looper.head_position
        2
        >>> looper.end_with_max_n_leaves
        False
        >>> looper.omit_time_signatures
        True
        >>> looper.process_on_first_call
        False
        >>> looper.after_rest
        abjad.Duration(0)
        >>> looper.after_rest_in_new_measure
        False
        >>> looper.use_multimeasure_rests
        True

    Setting :attr:`forward_bias` to ``0.0``:
        Set :attr:`forward_bias` to ``0.0`` to move backwards instead of
        forwards (default is ``1.0``). The initial :attr:`head_position` must
        be greater than ``0`` otherwise the contents will already be exhausted
        in the very first call (since it will not be able to move backwards
        from that position).

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=2,
        ...                            head_position=2,
        ...                            forward_bias=0.0,
        ...                            )
        >>> notes = looper.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                e'4
                f'4
                d'4
                e'4
                c'4
                d'4
            }

        ..  figure:: ../_images/LeafLooper-tdypneg8a1i.png

    :attr:`forward_bias` between ``0.0`` and ``1.0``:
        Setting :attr:`forward_bias` to a value in between ``0.0`` and ``1.0``
        will result in random steps being taken forward or backward, according
        to the bias. The initial value of :attr:`head_position` will once gain
        play an important role here, as the contents might be exhausted if the
        looper attempts to move backwards after reaching the head position
        ``0``.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4 b'4 c''4")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            head_position=3,
        ...                            forward_bias=0.5,
        ...                            )
        >>> notes = looper.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                f'4
                g'4
                a'4
                e'4
                f'4
                g'4
                f'4
                g'4
                a'4
                e'4
                f'4
                g'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/LeafLooper-la5quq8s3al.png

    :attr:`max_steps`:
        Setting the keyword argument :attr:`max_steps` to a value larger than
        ``1`` will result in a random number of steps (between ``1`` and
        :attr:`max_steps`) being applied at each call.

        >>> container = abjad.Container(
        ...     r"c'4 d'4 e'4 f'4 g'4 a'4 b'4 c''4 d''4 e''4"
        ... )
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=2,
        ...                            max_steps=4,
        ...                            )
        >>> notes = looper.output_n(4)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'4
                d'4
                g'4
                a'4
                a'4
                b'4
                c''4
                d''4
            }

        ..  figure:: ../_images/LeafLooper-u7t1hsb1o79.png

    :func:`len()`:
        The function :func:`len()` can be used to get the total number of
        elements in the contents.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            )
        >>> len(looper)
        5

    :meth:`output_all`:
        To run through the whole process and output it as a single container,
        from the initial head position until the process outputs the single
        last element, use the method :meth:`output_all`.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=2,
        ...                            )
        >>> window = looper.output_all()
        >>> staff = abjad.Staff(window)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'4
                d'4
                \time 2/4
                d'4
                e'4
                \time 2/4
                e'4
                f'4
                \time 1/4
                f'4
            }

        ..  figure:: ../_images/LeafLooper-o5yqw4xinp.png

    ``tie_identical_pitches``:
        When using :meth:`output_all`, set the keyword argument
        ``tie_identical_pitches`` to ``True`` in order to tie identical notes
        or chords at the end and beginning of consecutive windows.

        >>> container = abjad.Container(
        ...     r"c'4 d'2 r8 d'4 <e' g'>8 r4 f'2. <e' g'>16"
        ... )
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=4,
        ...                            disable_rewrite_meter=True,
        ...                            )
        >>> notes = looper.output_all(tie_identical_pitches=True)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 9/8
                c'4
                d'2
                r8
                d'4
                ~
                \time 4/4
                d'2
                r8
                d'4
                <e' g'>8
                \time 3/4
                r8
                d'4
                <e' g'>8
                r4
                \time 11/8
                d'4
                <e' g'>8
                r4
                f'2.
                \time 19/16
                <e' g'>8
                r4
                f'2.
                <e' g'>16
                \time 17/16
                r4
                f'2.
                <e' g'>16
                \time 13/16
                f'2.
                <e' g'>16
                ~
                \time 1/16
                <e' g'>16
            }

        ..  figure:: ../_images/LeafLooper-v1a75ntci7r.png

    :meth:`output_n`:
        To run through just part of the process and output it as a single
        container, starting from the initial head position, use the method
        :meth:`output_n` and pass the number of iterations as argument.
        Similarly to :meth:`output_all`, the keyword argument
        ``tie_identical_pitches`` is available for tying pitches.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=2,
        ...                            )
        >>> window = looper.output_n(2)
        >>> staff = abjad.Staff(window)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'4
                d'4
                \time 2/4
                d'4
                e'4
            }

        ..  figure:: ../_images/LeafLooper-w389c0wnl2.png

    ..  tip::

        All methods that return an |abjad.Selection| will add an initial time
        signature to it. The :meth:`output_n` and :meth:`output_all` methods
        automatically remove repeated time signatures. When joining selections
        output by multiple method calls, use
        |auxjad.mutate.remove_repeated_time_signatures()| on the whole
        container after fusing the selections to remove any unecessary time
        signature changes.

    :attr:`omit_time_signatures`:
        To disable time signatures altogether, initialise this class with the
        keyword argument :attr:`omit_time_signatures` set to ``True`` (default
        is ``False``), or use the :attr:`omit_time_signatures` property after
        initialisation.

        >>> container = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'4.")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            omit_time_signatures=True,
        ...                            )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'2
                e'4
            }

        ..  figure:: ../_images/LeafLooper-t3wqox0d0qe.png

    :attr:`after_rest`:
        To append rests after the output (useful to create separations between
        consecutive windows), set the keyword argument :attr:`after_rest` to
        an :obj:`int`, :obj:`float`, :obj:`tuple`, or |abjad.Duration| with the
        total length of the rests. Setting it to ``abjad.Duration(0)`` will
        disable rests (which is the default value).

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=2,
        ...                            after_rest=(1, 8),
        ...                            )
        >>> notes = looper.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 5/8
                c'4
                d'8
                ~
                d'8
                r8
                d'4
                e'8
                ~
                e'8
                r8
                e'4
                f'8
                ~
                f'8
                r8
                \time 3/8
                f'4
                r8
            }

        ..  figure:: ../_images/LeafLooper-AXXMUqnFfD.png

    :attr:`after_rest_in_new_measure`:
        When using after rests, the keyword argument
        :attr:`after_rest_in_new_measure` controls whether or not the rests are
        appended to the same output measure or to a new one.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=2,
        ...                            after_rest=(1, 8),
        ...                            after_rest_in_new_measure=True,
        ...                            )
        >>> notes = looper.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'4
                d'4
                \time 1/8
                R1 * 1/8
                \time 2/4
                d'4
                e'4
                \time 1/8
                R1 * 1/8
                \time 2/4
                e'4
                f'4
                \time 1/8
                R1 * 1/8
                \time 1/4
                f'4
                \time 1/8
                R1 * 1/8
            }

        ..  figure:: ../_images/LeafLooper-FQtegfLdcr.png


    :attr:`use_multimeasure_rests`:
        When after rests are used and :attr:`after_rest_in_new_measure` is set
        to ``True``, multi-measure rests are automatically used. To use regular
        rests, set :attr:`use_multimeasure_rests` to ``False``.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=2,
        ...                            after_rest=(1, 8),
        ...                            after_rest_in_new_measure=True,
        ...                            use_multimeasure_rests=False,
        ...                            )
        >>> notes = looper.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'4
                d'4
                \time 1/8
                r8
                \time 2/4
                d'4
                e'4
                \time 1/8
                r8
                \time 2/4
                e'4
                f'4
                \time 1/8
                r8
                \time 1/4
                f'4
                \time 1/8
                r8
            }

        ..  figure:: ../_images/LeafLooper-wqN3GAcMTR.png

    ..  note::

        Multi-measure rests will not be used if :attr:`omit_time_signatures` is
        ``True``, regardless of the value of :attr:`use_multimeasure_rests`.

    :attr:`end_with_max_n_leaves`:
        When ``True``, the last bar in the output will contain the maximum
        number of leaves given by :attr:`window_size`. E.g. consider the
        logical ties :math:`[A, B, C, D]` (where each letter represents one
        logical tie) and the looping window was size ``3``; setting
        :attr:`end_with_max_n_leaves` to ``True`` will output:

        :math:`A~ B~ C~ |~ B~ C~ D`

        Setting it to ``False`` (which is this property's default value) will
        produces:

        :math:`A~ B~ C~ |~ B~ C~ D~ |~ C~ D~ |~ D`

        Compare the two examples below:

        >>> container = abjad.Container(r"c'4 d'4. e'4 f'8")
        >>> looper = auxjad.LeafLooper(container, window_size=3)
        >>> notes = looper.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 7/8
                c'4
                d'4.
                e'4
                \time 3/4
                d'4.
                e'4
                f'8
                \time 3/8
                e'4
                f'8
                \time 1/8
                f'8
            }

        ..  figure:: ../_images/LeafLooper-aZUekjEPZZ.png

        >>> container = abjad.Container(r"c'4 d'4. e'4 f'8")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            end_with_max_n_leaves=True,
        ...                            )
        >>> notes = looper.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 7/8
                c'4
                d'4.
                e'4
                \time 3/4
                d'4.
                e'4
                f'8
            }

        ..  figure:: ../_images/LeafLooper-QYvLl9Ce9c.png

    :attr:`window_size`:
        To change the size of the looping window after instantiation, use the
        property :attr:`window_size`. In the example below, the initial window
        is of size ``3``, and so the first call of the looper object outputs
        the first, second, and third leaves. The window size is then set to
        ``4``, and the looper is called again, moving to the leaf in the next
        position, thus outputting the second, third, fourth, and fifth leaves.

        >>> container = abjad.Container(r"c'4 d'4 e'8 f'4 g'2")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 5/8
                c'4
                d'8
                ~
                d'8
                e'8
            }

        ..  figure:: ../_images/LeafLooper-euq7xez3whk.png

        >>> looper.window_size = 4
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 9/8
                d'4
                e'8
                f'4
                g'2
            }

        ..  figure:: ../_images/LeafLooper-9niipiwpbz.png

    :attr:`contents`:
        Use the :attr:`contents` property to read as well as overwrite the
        contents of the looper. Notice that the :attr:`head_position` will
        remain on its previous value and must be reset to ``0`` if that's
        required.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/LeafLooper-39uewm1fzrj.png

        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/LeafLooper-n6jrizfcon.png

        >>> looper.contents = abjad.Container(
        ...     r"cs'''4 ds'''4 es'''4 fs'''4"
        ... )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                ds'''4
                es'''4
                fs'''4
            }

        ..  figure:: ../_images/LeafLooper-3lqsqrd3sar.png

        >>> looper.head_position = 0
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                cs'''4
                ds'''4
                es'''4
            }

        ..  figure:: ../_images/LeafLooper-g3xybel6xu.png

    Indicators:
        This class supports dynamics and slurs.

        >>> container = abjad.Container(
        ...     r"c'4\p( d'2 e'4\f) f'2( ~ f'8 g'4 a'1\pp)"
        ... )
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            disable_rewrite_meter=True,
        ...                            )
        >>> notes = looper.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \p
                (
                d'2
                e'4
                \f
                )
                \time 11/8
                d'2
                \p
                (
                e'4
                \f
                )
                f'2
                ~
                f'8
                \time 9/8
                e'4
                f'2
                (
                ~
                f'8
                g'4
                )
                \time 15/8
                f'2
                (
                ~
                f'8
                g'4
                a'1
                \pp
                )
                \time 5/4
                g'4
                \f
                (
                a'1
                \pp
                )
            }

        ..  figure:: ../_images/LeafLooper-cqzbiawxp4.png

    :attr:`disable_rewrite_meter`:
        By default, this class uses the |abjad.Meter.rewrite_meter()|
        mutation.

        >>> container = abjad.Container(r"c'16 d'4 e'8. f'4 g'16")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            )
        >>> notes = looper.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'16
                d'8.
                ~
                d'16
                e'8.
                \time 11/16
                d'4
                e'8.
                f'4
                \time 2/4
                e'8.
                f'16
                ~
                f'8.
                g'16
            }

        ..  figure:: ../_images/LeafLooper-ab18rc3voqn.png

        Set :attr:`disable_rewrite_meter` to ``True`` in order to disable this
        behaviour.

        >>> container = abjad.Container(r"c'16 d'4 e'8. f'4 g'16")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=3,
        ...                            disable_rewrite_meter=True,
        ...                            )
        >>> notes = looper.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'16
                d'4
                e'8.
                \time 11/16
                d'4
                e'8.
                f'4
                \time 2/4
                e'8.
                f'4
                g'16
            }

        ..  figure:: ../_images/LeafLooper-fa30crzy146.png

    ..  note::

        This class also accepts the properties ``boundary_depth``,
        ``maximum_dot_count``, and ``rewrite_tuplets``, which are passed on to
        |abjad.Meter.rewrite_meter()|, and ``fuse_across_groups_of_beats``,
        ``fuse_quadruple_meter``, ``fuse_triple_meter``, and
        ``extract_trivial_tuplets``, which are passed on to
        |auxjad.mutate.prettify_rewrite_meter()| (the latter can be disabled
        by setting ``prettify_rewrite_meter`` to ``False``). See the
        documentation of those functions for more details on these arguments.

    ..  warning::

        This class can handle tuplets, but the engraving of the output is often
        quite complex and thus this functionality should be considered as
        experimental. Time signatures will be correct when dealing with partial
        tuplets (thus having non-standard values in their denominators) and
        consecutive partial tuplets will be merged when possible.

        >>> container = abjad.Container(r"c'4 d'8 \times 2/3 {a4 g2}")
        >>> looper = auxjad.LeafLooper(container,
        ...                            window_size=2,
        ...                            )
        >>> window = looper.output_all()
        >>> staff = abjad.Staff(window)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4
                d'8
                #(ly:expect-warning "strange time signature found")
                \time 7/24
                d'8
                \tweak edge-height #'(0.7 . 0)
                \times 2/3
                {
                    a4
                }
                \times 2/3
                {
                    \time 2/4
                    a4
                    g2
                }
                \tweak edge-height #'(0.7 . 0)
                \times 2/3
                {
                    #(ly:expect-warning "strange time signature found")
                    \time 2/6
                    g2
                }
            }

        ..  figure:: ../_images/LeafLooper-nre2ecw27l.png

    ..  tip::

        The functions |auxjad.mutate.remove_repeated_dynamics()| and
        |auxjad.mutate.reposition_clefs()| can be used to clean the output
        and remove repeated dynamics and unnecessary clef changes.

    ..  warning::

        Do note that some elements that span multiple notes (such as ottava
        indicators, manual beams, etc.) can become problematic when notes
        containing them are split into two. As a rule of thumb, it is always
        better to attach those to the music after the fading process has ended.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_end_with_max_n_leaves",
        "_omit_time_signatures",
        "_contents_logical_ties",
        "_disable_rewrite_meter",
        "_boundary_depth",
        "_maximum_dot_count",
        "_rewrite_tuplets",
        "_prettify_rewrite_meter",
        "_extract_trivial_tuplets",
        "_fuse_across_groups_of_beats",
        "_fuse_quadruple_meter",
        "_fuse_triple_meter",
        "_after_rest",
        "_after_rest_in_new_measure",
        "_use_multimeasure_rests",
    )

    ### INITIALISER ###

    def __init__(
        self,
        contents: abjad.Container,
        *,
        window_size: int,
        step_size: int = 1,
        max_steps: int = 1,
        repetition_chance: float = 0.0,
        forward_bias: float = 1.0,
        head_position: int = 0,
        end_with_max_n_leaves: bool = False,
        omit_time_signatures: bool = False,
        process_on_first_call: bool = False,
        disable_rewrite_meter: bool = False,
        boundary_depth: Optional[int] = None,
        maximum_dot_count: Optional[int] = None,
        rewrite_tuplets: bool = True,
        prettify_rewrite_meter: bool = True,
        extract_trivial_tuplets: bool = True,
        fuse_across_groups_of_beats: bool = True,
        fuse_quadruple_meter: bool = True,
        fuse_triple_meter: bool = True,
        after_rest: Union[
            int,
            float,
            str,
            tuple[int],
            abjad.Duration,
            abjad.Rest,
        ] = 0,
        after_rest_in_new_measure: bool = False,
        use_multimeasure_rests: bool = True,
    ) -> None:
        self.contents = contents
        self.end_with_max_n_leaves = end_with_max_n_leaves
        self.omit_time_signatures = omit_time_signatures
        self.disable_rewrite_meter = disable_rewrite_meter
        self.boundary_depth = boundary_depth
        self.maximum_dot_count = maximum_dot_count
        self.rewrite_tuplets = rewrite_tuplets
        self.prettify_rewrite_meter = prettify_rewrite_meter
        self.extract_trivial_tuplets = extract_trivial_tuplets
        self.fuse_across_groups_of_beats = fuse_across_groups_of_beats
        self.fuse_quadruple_meter = fuse_quadruple_meter
        self.fuse_triple_meter = fuse_triple_meter
        self.after_rest = after_rest
        self.after_rest_in_new_measure = after_rest_in_new_measure
        self.use_multimeasure_rests = use_multimeasure_rests
        super().__init__(
            head_position=head_position,
            window_size=window_size,
            step_size=step_size,
            max_steps=max_steps,
            repetition_chance=repetition_chance,
            forward_bias=forward_bias,
            process_on_first_call=process_on_first_call,
        )

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of :attr:`contents`."""
        return abjad.lilypond(self._contents)

    def __len__(self) -> int:
        r"""Returns the number of logical ties of :attr:`contents`."""
        return len(self._contents_logical_ties)

    ### PRIVATE METHODS ###

    def _slice_contents(self) -> None:
        r"""This method takes a slice with :attr:`window_size` number of
        logical ties out of :attr:`contents` starting at the current
        :attr:`head_position`.
        """
        start = self._head_position
        end = self._head_position + self._window_size
        logical_ties = self._contents_logical_ties[start:end]
        dummy_container = abjad.Container()
        time_signature_duration = 0
        for logical_tie in logical_ties:
            effective_duration = abjad.get.duration(logical_tie)
            logical_tie_ = abjad.mutate.copy(logical_tie)
            dummy_container.append(logical_tie_)
            multiplier = effective_duration / logical_tie_.written_duration
            try:
                logical_tie_ = abjad.mutate.scale(logical_tie_, multiplier)
            except abjad.exceptions.AssignabilityError:
                tuplet = abjad.Tuplet(
                    multiplier,
                    abjad.mutate.copy(logical_tie),
                )
                abjad.mutate.replace(logical_tie_, tuplet)
            time_signature_duration += effective_duration
        if self._after_rest > 0:
            if (
                self._after_rest_in_new_measure
                and self._use_multimeasure_rests
                and not self._omit_time_signatures
            ):
                if self._after_rest == 1:
                    multiplier = None
                else:
                    multiplier = abjad.Multiplier(self._after_rest)
                rest = abjad.MultimeasureRest(
                    (4, 4),
                    multiplier=multiplier,
                )
            else:
                rest = abjad.LeafMaker()([None], [self._after_rest])
            if not self._after_rest_in_new_measure:
                time_signature_duration += self._after_rest
            else:
                time_signature = abjad.TimeSignature(self._after_rest)
                time_signature.simplify_ratio()
                abjad.attach(time_signature, abjad.select(rest).leaf(0))
            dummy_container.append(rest)
        if len(logical_ties) > 0:
            time_signature = abjad.TimeSignature(time_signature_duration)
            time_signature.simplify_ratio()
            abjad.attach(time_signature, abjad.select(dummy_container).leaf(0))
        self._notate_music(dummy_container, start)

    def _notate_music(
        self,
        dummy_container: abjad.Container,
        start: int,
    ) -> None:
        r"""Handles the notation aspects of the looping window."""
        start_head = abjad.select(dummy_container).logical_tie(0)[0]
        start_tail = abjad.select(dummy_container).logical_tie(0)[-1]
        if (
            abjad.get.indicator(start_head, abjad.StartSlur) is None
            and abjad.get.indicator(start_tail, abjad.StopSlur) is None
        ):
            for leaf in self._contents_logical_ties[start - 1 :: -1].leaves():
                if abjad.get.indicator(leaf, abjad.StartSlur) is not None:
                    abjad.attach(abjad.StartSlur(), start_head)
                    break
                elif abjad.get.indicator(leaf, abjad.StopSlur) is not None:
                    break
        if abjad.get.indicator(start_head, abjad.Dynamic) is None:
            for leaf in self._contents_logical_ties[start - 1 :: -1].leaves():
                dynamic = abjad.get.indicator(leaf, abjad.Dynamic)
                if dynamic is not None:
                    abjad.attach(dynamic, start_head)
                    break
        mutate.reposition_dynamics(dummy_container[:])
        mutate.reposition_slurs(dummy_container[:])
        # rewriting meter
        if not self._disable_rewrite_meter:
            mutate.auto_rewrite_meter(
                dummy_container,
                boundary_depth=self._boundary_depth,
                maximum_dot_count=self._maximum_dot_count,
                rewrite_tuplets=self._rewrite_tuplets,
                prettify_rewrite_meter=self._prettify_rewrite_meter,
                extract_trivial_tuplets=self._extract_trivial_tuplets,
                fuse_across_groups_of_beats=self._fuse_across_groups_of_beats,
                fuse_quadruple_meter=self._fuse_quadruple_meter,
                fuse_triple_meter=self._fuse_triple_meter,
            )
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    def _get_lilypond_format(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return self.__repr__()

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> abjad.Container:
        r"""The |abjad.Container| to be sliced and looped."""
        return abjad.mutate.copy(self._contents)

    @contents.setter
    def contents(
        self,
        contents: abjad.Container,
    ) -> None:
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or child class")
        if not abjad.select(contents).leaves().are_contiguous_logical_voice():
            raise ValueError("'contents' must be contiguous logical voice")
        if isinstance(contents, abjad.Score):
            self._contents = abjad.mutate.copy(contents[0])
        elif isinstance(contents, abjad.Tuplet):
            self._contents = abjad.Container([abjad.mutate.copy(contents)])
        else:
            self._contents = abjad.mutate.copy(contents)
        dummy_container = abjad.mutate.copy(self._contents)
        self._remove_all_time_signatures(dummy_container)
        selector = abjad.select(dummy_container)
        self._contents_logical_ties = selector.logical_ties()
        self._is_first_window = True

    @property
    def end_with_max_n_leaves(self) -> bool:
        r"""When ``True``, the last bar in the output will contain the maximum
        number of leaves given by :attr:`window_size`. E.g. consider the
        logical ties :math:`[A, B, C, D]` (where each letter represents one
        logical tie) and the looping window was size ``3``; setting
        :attr:`end_with_max_n_leaves` to ``True`` will output:

        :math:`A~ B~ C~ |~ B~ C~ D`

        Setting it to ``False`` (which is this property's default value) will
        produces:

        :math:`A~ B~ C~ |~ B~ C~ D~ |~ C~ D~ |~ D`
        """
        return self._end_with_max_n_leaves

    @end_with_max_n_leaves.setter
    def end_with_max_n_leaves(
        self,
        end_with_max_n_leaves: bool,
    ) -> None:
        if not isinstance(end_with_max_n_leaves, bool):
            raise TypeError("'end_with_max_n_leaves' must be 'bool'")
        self._end_with_max_n_leaves = end_with_max_n_leaves

    @property
    def omit_time_signatures(self) -> bool:
        r"""When ``True``, the output will contain no time signatures."""
        return self._omit_time_signatures

    @omit_time_signatures.setter
    def omit_time_signatures(
        self,
        omit_time_signatures: bool,
    ) -> None:
        if not isinstance(omit_time_signatures, bool):
            raise TypeError("'omit_time_signatures' must be 'bool'")
        self._omit_time_signatures = omit_time_signatures

    @property
    def disable_rewrite_meter(self) -> bool:
        r"""When ``True``, the durations of the notes in the output will not be
        rewritten by the |abjad.Meter.rewrite_meter()| mutation. Rests will
        have the same duration as the logical ties they replaced.
        """
        return self._disable_rewrite_meter

    @disable_rewrite_meter.setter
    def disable_rewrite_meter(
        self,
        disable_rewrite_meter: bool,
    ) -> None:
        if not isinstance(disable_rewrite_meter, bool):
            raise TypeError("'disable_rewrite_meter' must be 'bool'")
        self._disable_rewrite_meter = disable_rewrite_meter

    @property
    def boundary_depth(self) -> Union[int, None]:
        r"""Sets the argument ``boundary_depth`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._boundary_depth

    @boundary_depth.setter
    def boundary_depth(
        self,
        boundary_depth: Optional[int],
    ) -> None:
        if boundary_depth is not None:
            if not isinstance(boundary_depth, int):
                raise TypeError("'boundary_depth' must be 'int'")
        self._boundary_depth = boundary_depth

    @property
    def maximum_dot_count(self) -> Union[int, None]:
        r"""Sets the argument ``maximum_dot_count`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._maximum_dot_count

    @maximum_dot_count.setter
    def maximum_dot_count(
        self,
        maximum_dot_count: Optional[int],
    ) -> None:
        if maximum_dot_count is not None:
            if not isinstance(maximum_dot_count, int):
                raise TypeError("'maximum_dot_count' must be 'int'")
        self._maximum_dot_count = maximum_dot_count

    @property
    def rewrite_tuplets(self) -> bool:
        r"""Sets the argument ``rewrite_tuplets`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._rewrite_tuplets

    @rewrite_tuplets.setter
    def rewrite_tuplets(
        self,
        rewrite_tuplets: bool,
    ) -> None:
        if not isinstance(rewrite_tuplets, bool):
            raise TypeError("'rewrite_tuplets' must be 'bool'")
        self._rewrite_tuplets = rewrite_tuplets

    @property
    def prettify_rewrite_meter(self) -> bool:
        r"""Used to enable or disable the mutation
        |auxjad.mutate.prettify_rewrite_meter()| (default ``True``).
        """
        return self._prettify_rewrite_meter

    @prettify_rewrite_meter.setter
    def prettify_rewrite_meter(
        self,
        prettify_rewrite_meter: bool,
    ) -> None:
        if not isinstance(prettify_rewrite_meter, bool):
            raise TypeError("'prettify_rewrite_meter' must be 'bool'")
        self._prettify_rewrite_meter = prettify_rewrite_meter

    @property
    def extract_trivial_tuplets(self) -> bool:
        r"""Sets the argument ``extract_trivial_tuplets`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._extract_trivial_tuplets

    @extract_trivial_tuplets.setter
    def extract_trivial_tuplets(
        self,
        extract_trivial_tuplets: bool,
    ) -> None:
        if not isinstance(extract_trivial_tuplets, bool):
            raise TypeError("'extract_trivial_tuplets' must be 'bool'")
        self._extract_trivial_tuplets = extract_trivial_tuplets

    @property
    def fuse_across_groups_of_beats(self) -> bool:
        r"""Sets the argument ``fuse_across_groups_of_beats`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_across_groups_of_beats

    @fuse_across_groups_of_beats.setter
    def fuse_across_groups_of_beats(
        self,
        fuse_across_groups_of_beats: bool,
    ) -> None:
        if not isinstance(fuse_across_groups_of_beats, bool):
            raise TypeError("'fuse_across_groups_of_beats' must be 'bool'")
        self._fuse_across_groups_of_beats = fuse_across_groups_of_beats

    @property
    def fuse_quadruple_meter(self) -> bool:
        r"""Sets the argument ``fuse_quadruple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_quadruple_meter

    @fuse_quadruple_meter.setter
    def fuse_quadruple_meter(
        self,
        fuse_quadruple_meter: bool,
    ) -> None:
        if not isinstance(fuse_quadruple_meter, bool):
            raise TypeError("'fuse_quadruple_meter' must be 'bool'")
        self._fuse_quadruple_meter = fuse_quadruple_meter

    @property
    def fuse_triple_meter(self) -> bool:
        r"""Sets the argument ``fuse_triple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_triple_meter

    @fuse_triple_meter.setter
    def fuse_triple_meter(
        self,
        fuse_triple_meter: bool,
    ) -> None:
        if not isinstance(fuse_triple_meter, bool):
            raise TypeError("'fuse_triple_meter' must be 'bool'")
        self._fuse_triple_meter = fuse_triple_meter

    @property
    def after_rest(self) -> abjad.Duration:
        r"""Sets the length of the rest appended at the end of the window
        (default is ``0``).
        """
        return self._after_rest

    @after_rest.setter
    def after_rest(
        self,
        after_rest: Union[
            int,
            float,
            str,
            tuple[int],
            abjad.Duration,
            abjad.Rest,
        ],
    ) -> None:
        if not isinstance(
            after_rest,
            (abjad.Duration, abjad.Rest, str, tuple, int, float),
        ):
            raise TypeError(
                "'after_rest' must be 'abjad.Duration', "
                "'abjad.Rest', 'str', 'tuple', or a number"
            )
        if isinstance(after_rest, abjad.Rest):
            after_rest = abjad.get.duration(after_rest)
        self._after_rest = abjad.Duration(after_rest)

    @property
    def after_rest_in_new_measure(self) -> bool:
        r"""If ``True``, then after rests will be added to their own measure
        (default is ``False``).
        """
        return self._after_rest_in_new_measure

    @after_rest_in_new_measure.setter
    def after_rest_in_new_measure(
        self,
        after_rest_in_new_measure: bool,
    ) -> None:
        if not isinstance(after_rest_in_new_measure, bool):
            raise TypeError("'after_rest_in_new_measure' must be 'bool'")
        self._after_rest_in_new_measure = after_rest_in_new_measure

    @property
    def use_multimeasure_rests(self) -> bool:
        r"""If ``True``, then multi-measure rests will be used for after rests
        when added to their own measure (default is ``True``).
        """
        return self._use_multimeasure_rests

    @use_multimeasure_rests.setter
    def use_multimeasure_rests(
        self,
        use_multimeasure_rests: bool,
    ) -> None:
        if not isinstance(use_multimeasure_rests, bool):
            raise TypeError("'use_multimeasure_rests' must be 'bool'")
        self._use_multimeasure_rests = use_multimeasure_rests

    ### PRIVATE PROPERTIES ###

    @property
    def _done(self) -> bool:
        r""":obj:`bool` indicating whether the process is done (i.e. whether
        the head position has overtaken the :attr:`contents`'s length).
        """
        if self._end_with_max_n_leaves:
            return (
                self._head_position + self._window_size > self.__len__() or self._head_position < 0
            )
        else:
            return self._head_position >= self.__len__() or self._head_position < 0



import abjad


def respell_augmented_unisons(
    selection: abjad.Selection,
    *,
    include_multiples: bool = False,
    respell_by_pitch_class: bool = False,
) -> None:
    r"""Mutates an input |abjad.Selection| in place and has no return value;
    this function changes the accidentals of individual pitches of all chords
    in a container in order to avoid augmented unisons.

    Basic usage:
        To use this function, apply it to a selection that contains chords that
        have augmented unisons.

        >>> container = abjad.Container(r"c'4 r4 <ef' e'>4 g'4 <c' cs'>4 r2.")
        >>> abjad.show(container)

        ..  docs::

            {
                c'4
                r4
                <ef' e'>4
                g'4
                <c' cs'>4
                r2.
            }

        ..  figure:: ../_images/respell_augmented_unisons-OXnGvQzGT2.png

        >>> auxjad.mutate.respell_augmented_unisons(container[:])
        >>> abjad.show(container)

        ..  docs::

            {
                c'4
                r4
                <ds' e'>4
                g'4
                <c' df'>4
                r2.
            }

        ..  figure:: ../_images/respell_augmented_unisons-x33afbbamt.png

        This can be useful when using tuples of integers to create chords that
        contain minor seconds:

        >>> pitches = [(0, 1), (8, 9, 12), (0, 4, 5, 6), (-1, 5, 6)]
        >>> durations = [(1, 8), (3, 8), (7, 16), (1, 16)]
        >>> maker = abjad.LeafMaker()
        >>> chords = maker(pitches, durations)
        >>> staff = abjad.Staff(chords)
        >>> literal = abjad.LilyPondLiteral(r'\accidentalStyle dodecaphonic')
        >>> abjad.attach(literal, staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \accidentalStyle dodecaphonic
                <c' cs'>8
                <af' a' c''>4.
                <c' e' f' fs'>4..
                <b f' fs'>16
            }

        ..  figure:: ../_images/respell_augmented_unisons-mWVWwV9uBx.png

        >>> auxjad.mutate.respell_augmented_unisons(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \accidentalStyle dodecaphonic
                <c' df'>8
                <gs' a' c''>4.
                <c' e' f' gf'>4..
                <b f' gf'>16
            }

        ..  figure:: ../_images/respell_augmented_unisons-IfzaseW4oS.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.respell_augmented_unisons(staff[:])
        >>> abjad.mutate.respell_augmented_unisons(staff[:])

    2-note chords:
        The example below shows first the default spelling of 2-note chords in
        Abjad followed by the respelt 2-note chords.

        >>> staff = abjad.Staff()
        >>> for pitch in range(12):
        ...     staff.append(abjad.Chord([pitch, pitch + 1], (1, 16)))
        >>> literal = abjad.LilyPondLiteral(r'\accidentalStyle dodecaphonic')
        >>> abjad.attach(literal, staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \accidentalStyle dodecaphonic
                <c' cs'>16
                <cs' d'>16
                <d' ef'>16
                <ef' e'>16
                <e' f'>16
                <f' fs'>16
                <fs' g'>16
                <g' af'>16
                <af' a'>16
                <a' bf'>16
                <bf' b'>16
                <b' c''>16
            }

        ..  figure:: ../_images/respell_augmented_unisons-qQduIqCRpz.png

        >>> auxjad.mutate.respell_augmented_unisons(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \accidentalStyle dodecaphonic
                <c' df'>16
                <cs' d'>16
                <d' ef'>16
                <ds' e'>16
                <e' f'>16
                <f' gf'>16
                <fs' g'>16
                <g' af'>16
                <gs' a'>16
                <a' bf'>16
                <as' b'>16
                <b' c''>16
            }

        ..  figure:: ../_images/respell_augmented_unisons-jvg032q24il.png

    augmented unissons in chords with 3 or more pitches:
        The function looks for all augmented unissons in chords of 3 or more
        pitches:

        >>> staff = abjad.Staff(r"<a c' cs' f'>1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <a c' cs' f'>1
            }

        ..  figure:: ../_images/respell_augmented_unisons-IklJO81q1E.png

        >>> auxjad.mutate.respell_augmented_unisons(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <a c' df' f'>1
            }

        ..  figure:: ../_images/respell_augmented_unisons-gyficck05p.png

    ``include_multiples``:
        By default, this function only changes spelling for pitches that are
        1 semitone apart.

        >>> staff = abjad.Staff(r"<c' cs''>1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <c' cs''>1
            }

        ..  figure:: ../_images/respell_augmented_unisons-HPqFrADjeh.png

        >>> auxjad.mutate.respell_augmented_unisons(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <c' cs''>1
            }

        ..  figure:: ../_images/respell_augmented_unisons-uszf11qb72d.png

        To consider pitches in different octaves (thus including augmented
        unisons, augmented octaves, augmented fifteenths, etc.), call this
        function with the keyword argument ``include_multiples`` set to
        ``True``.

        >>> staff = abjad.Staff(r"<c' cs''>1")
        >>> auxjad.mutate.respell_augmented_unisons(
        ...     staff[:],
        ...     include_multiples=True,
        ... )

        ..  docs::

            \new Staff
            {
                <c' df''>1
            }

        ..  figure:: ../_images/respell_augmented_unisons-8am8cu2rmgi.png

    ``respell_by_pitch_class``:
        By default, when this function changes the spelling of a pitch, it does
        not change the spelling of all other pitches with the same pitch-class.

        >>> staff = abjad.Staff(r"<c' cs' cs''>1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                    <c' cs' cs''>1
            }

        ..  figure:: ../_images/respell_augmented_unisons-eWixL7iCEq.png

        >>> auxjad.mutate.respell_augmented_unisons(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <c' df' cs''>1
            }

        ..  figure:: ../_images/respell_augmented_unisons-47d16xk6gvs.png

        To alter all pitch-classes, call this function with the keyword
        argument ``respell_by_pitch_class`` set to ``True``.

        >>> staff = abjad.Staff(r"<c' cs' cs''>1")
        >>> auxjad.mutate.respell_augmented_unisons(
        ...     staff[:],
        ...     respell_by_pitch_class=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <c' df' df''>1
            }

        ..  figure:: ../_images/respell_augmented_unisons-kobft0oq9sl.png
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Selection'")
    if not isinstance(respell_by_pitch_class, bool):
        raise TypeError("'respell_by_pitch_class' must be 'bool'")
    if not isinstance(include_multiples, bool):
        raise TypeError("'include_multiples' must be 'bool'")

    for leaf in selection.leaves():
        if isinstance(leaf, abjad.Chord):
            accidentals = [pitch.accidental for pitch in leaf.written_pitches]
            original_accidentals = accidentals[:]
            if not include_multiples:
                # dealing only with intervals of size equal to 1 semitone
                for i in range(len(leaf.written_pitches) - 1):
                    p1 = leaf.written_pitches[i]
                    p2 = leaf.written_pitches[i + 1]
                    interval12 = p1 - p2
                    try:
                        p3 = leaf.written_pitches[i + 2]
                        interval23 = p2 - p3
                    except IndexError:
                        p3 = None
                        interval23 = None
                    if (
                        interval12 == abjad.NamedInterval("+A1")
                        and interval23 != abjad.NamedInterval("+A1")
                        and interval23 != abjad.NamedInterval("+m2")
                    ):
                        if not respell_by_pitch_class:
                            # respelling only one single note
                            if p1.accidental == abjad.Accidental("f"):
                                accidentals[i] = abjad.Accidental("s")
                            elif p1.accidental == abjad.Accidental(""):
                                accidentals[i + 1] = abjad.Accidental("f")
                        else:
                            # respelling all matching pitch-classes
                            if p1.accidental == abjad.Accidental("f"):
                                for j, p in enumerate(leaf.written_pitches):
                                    if p.pitch_class == p1.pitch_class:
                                        accidentals[j] = abjad.Accidental("s")
                            elif p1.accidental == abjad.Accidental(""):
                                for j, p in enumerate(leaf.written_pitches):
                                    if p.pitch_class == p2.pitch_class:
                                        accidentals[j] = abjad.Accidental("f")
            else:
                # dealing with augmented unisons as well as augmented 8as,
                # 15ths, etc.
                for i in range(len(leaf.written_pitches) - 1):
                    for j in range(i + 1, len(leaf.written_pitches)):
                        # no p3 this time since p1 and p2 are not necessary
                        # consecutive pitches in the chord
                        p1 = leaf.written_pitches[i]
                        p2 = leaf.written_pitches[j]
                        interval = abjad.NamedIntervalClass(p1 - p2)
                        if interval in (
                            abjad.NamedIntervalClass("+A1"),
                            abjad.NamedIntervalClass("-d1"),
                        ):
                            # no need for respell_by_pitch_class since this
                            # will go through all notes in the chord anyway
                            if p1.accidental == abjad.Accidental("f"):
                                accidentals[i] = abjad.Accidental("s")
                            if p1.accidental == abjad.Accidental("s"):
                                accidentals[i] = abjad.Accidental("f")
                            elif p1.accidental == abjad.Accidental(""):
                                if p2.accidental == abjad.Accidental("s"):
                                    accidentals[j] = abjad.Accidental("f")
                                elif p2.accidental == abjad.Accidental("f"):
                                    accidentals[j] = abjad.Accidental("s")
            # rewritting chord with new spelling
            respelt_pitches = []
            for pitch, accidental, original_accidental in zip(
                leaf.written_pitches,
                accidentals,
                original_accidentals,
            ):
                if accidental != original_accidental and accidental == abjad.Accidental("f"):
                    respelt_pitches.append(pitch._respell(accidental="flats"))
                elif accidental != original_accidental and accidental == abjad.Accidental("s"):
                    respelt_pitches.append(pitch._respell(accidental="sharps"))
                else:
                    respelt_pitches.append(pitch)
            leaf.written_pitches = respelt_pitches
